#!/usr/bin/python3

import os
import shutil
from typing import Dict, List

import click

from .common import common_options, namespace_option


def get_nomad_commands() -> Dict[str, str]:
    nomad_help = """\
        run         Run a new job or update an existing job
        stop        Stop a running job
        status      Display the status output for a resource
        alloc       Interact with allocations
        job         Interact with jobs
        node        Interact with nodes
        agent       Runs a Nomad agent
        acl                 Interact with ACL policies and tokens
        agent-info          Display status information about the local agent
        config              Interact with configurations
        deployment          Interact with deployments
        eval                Interact with evaluations
        exec                Execute commands in task
        fmt                 Rewrites Nomad config and job files to canonical format
        license             Interact with Nomad Enterprise License
        login               Login to Nomad using an auth method
        monitor             Stream logs from a Nomad agent
        namespace           Interact with namespaces
        operator            Provides cluster-level tools for Nomad operators
        plugin              Inspect plugins
        quota               Interact with quotas
        recommendation      Interact with the Nomad recommendation endpoint
        scaling             Interact with the Nomad scaling endpoint
        sentinel            Interact with Sentinel policies
        server              Interact with servers
        service             Interact with registered services
        system              Interact with the system API
        tls                 Generate Self Signed TLS Certificates for Nomad
        ui                  Open the Nomad Web UI
        var                 Interact with variables
        version             Prints the Nomad version
        volume              Interact with volumes
    """
    ret = {}
    for x in nomad_help.splitlines():
        if x:
            tmp = x.strip().split(" ", 1)
            if tmp and len(tmp) == 2:
                ret[tmp[0]] = tmp[1].strip()
    return ret


def get_additional_commands() -> Dict[str, str]:
    """
    Search all executables in PATH environment variable and returns a List[str] of executables starting with prefix.
    """
    path_env = os.environ.get("PATH", os.defpath)
    executables = []
    prefix = "nomad-"
    for path in path_env.split(os.pathsep):
        try:
            for file in os.listdir(path):
                if file.startswith(prefix):
                    executables.append(file.lstrip(prefix))
        except FileNotFoundError:
            pass
    return {exec: f"**Run {prefix}{exec}" for exec in executables}


commands: Dict[str, str] = {
    **get_nomad_commands(),
    **get_additional_commands(),
}

###############################################################################


@click.group(
    help=f"""
If 'nomad-command[0]' exists, run it.
Otherwise, run 'nomad' with the arguments.
""",
)
@namespace_option()
@common_options()
def cli(**kwargs):
    pass


for cmd, help in commands.items():

    @cli.command(
        cmd,
        help=help,
        context_settings=dict(
            ignore_unknown_options=True,
            allow_extra_args=True,
        ),
    )
    @click.argument("command", nargs=-1, type=click.UNPROCESSED)
    @click.pass_context
    def func(ctx: click.Context, command: List[str]):
        name = ctx.command.name
        assert name
        command = [name, *command]
        cmd = f"nomad-{command[0]}"
        if shutil.which(cmd):
            os.execvp(cmd, command)
        else:
            os.execvp("nomad", command)


if __name__ == "__main__":
    cli()



import json
from typing import Any, Dict


def _parse(name: str, data: Dict[Any, Any]):
    classes = []
    lines = []
    for key, val in data.items():
        line = ""
        if isinstance(val, dict):
            classes += _parse(f"{name}{key}", val)
            line = f"{key}: {name}{key}"
        elif isinstance(val, list):
            val = val[0]
            if any(isinstance(val, x) for x in [str, int]):
                line = f"{key}: List[{val.__class__.__name__}]"
            elif isinstance(val, dict):
                classes += _parse(f"{name}{key}", val)
                line = f"{key}: List[{name}{key}]"
        elif any(isinstance(val, x) for x in [str, int]):
            line = f"{key}: {val.__class__.__name__}"
        elif val is None:
            line = f"{key}: Optional[Any] = None"
        assert line, f"{key} {val} {type(val)}"
        lines += [f"  {line}"]
    if name:
        cl = "\n".join([f"class {name}:"] + lines)
        classes += [cl]
    return classes


def generate(name: str, data: Dict[Any, Any]):
    return "\n\n".join(_parse(name, data))


def main(name: str, txt: str):
    print(generate(name, json.loads(txt)))
    print()
    print()


if __name__ == "__main__":
    print(generate("Test", {"a": 1, "b": "str", "c": {"d": "str", "e": [2, 2, 3]}}))

    main(
        "Job",
        """
        {
    "Region": null,
    "Namespace": null,
    "ID": "example",
    "Name": "example",
    "Type": "service",
    "Priority": null,
    "AllAtOnce": null,
    "Datacenters": [
      "dc1"
    ],
    "NodePool": "prod",
    "Constraints": null,
    "Affinities": null,
    "TaskGroups": [
      {
        "Name": "cache",
        "Count": 1,
        "Constraints": null,
        "Affinities": null,
        "Tasks": [
          {
            "Name": "redis",
            "Driver": "docker",
            "User": "",
            "Lifecycle": null,
            "Config": {
              "auth_soft_fail": true,
              "image": "redis:7",
              "ports": [
                "db"
              ]
            },
            "Constraints": null,
            "Affinities": null,
            "Env": null,
            "Services": null,
            "Resources": {
              "CPU": 500,
              "Cores": null,
              "MemoryMB": 256,
              "MemoryMaxMB": null,
              "DiskMB": null,
              "Networks": null,
              "Devices": null,
              "IOPS": null
            },
            "RestartPolicy": null,
            "Meta": null,
            "KillTimeout": null,
            "LogConfig": null,
            "Artifacts": null,
            "Vault": null,
            "Templates": null,
            "DispatchPayload": null,
            "VolumeMounts": null,
            "Leader": false,
            "ShutdownDelay": 0,
            "KillSignal": "",
            "Kind": "",
            "ScalingPolicies": null,
            "Identity": {
              "Env": true,
              "File": true
            }
          }
        ],
        "Spreads": null,
        "Volumes": null,
        "RestartPolicy": {
          "Interval": 1800000000000,
          "Attempts": 2,
          "Delay": 15000000000,
          "Mode": "fail"
        },
        "ReschedulePolicy": null,
        "EphemeralDisk": {
          "Sticky": null,
          "Migrate": null,
          "SizeMB": 300
        },
        "Update": null,
        "Migrate": null,
        "Networks": [
          {
            "Mode": "",
            "Device": "",
            "CIDR": "",
            "IP": "",
            "DNS": null,
            "ReservedPorts": null,
            "DynamicPorts": [
              {
                "Label": "db",
                "Value": 0,
                "To": 6379,
                "HostNetwork": ""
              }
            ],
            "Hostname": "",
            "MBits": null
          }
        ],
        "Meta": null,
        "Services": [
          {
            "Name": "redis-cache",
            "Tags": [
              "global",
              "cache"
            ],
            "CanaryTags": null,
            "EnableTagOverride": false,
            "PortLabel": "db",
            "AddressMode": "",
            "Address": "",
            "Checks": null,
            "CheckRestart": null,
            "Connect": null,
            "Meta": null,
            "CanaryMeta": null,
            "TaggedAddresses": null,
            "TaskName": "",
            "OnUpdate": "",
            "Provider": "nomad"
          }
        ],
        "ShutdownDelay": null,
        "StopAfterClientDisconnect": null,
        "MaxClientDisconnect": null,
        "Scaling": null,
        "Consul": null
      }
    ],
    "Update": {
      "Stagger": null,
      "MaxParallel": 1,
      "HealthCheck": null,
      "MinHealthyTime": 10000000000,
      "HealthyDeadline": 180000000000,
      "ProgressDeadline": 600000000000,
      "Canary": 0,
      "AutoRevert": false,
      "AutoPromote": null
    },
    "Multiregion": null,
    "Spreads": null,
    "Periodic": null,
    "ParameterizedJob": null,
    "Reschedule": null,
    "Migrate": {
      "MaxParallel": 1,
      "HealthCheck": "checks",
      "MinHealthyTime": 10000000000,
      "HealthyDeadline": 300000000000
    },
    "Meta": null,
    "ConsulToken": null,
    "VaultToken": null,
    "Stop": null,
    "ParentID": null,
    "Dispatched": false,
    "DispatchIdempotencyToken": null,
    "Payload": null,
    "ConsulNamespace": null,
    "VaultNamespace": null,
    "NomadTokenID": null,
    "Status": null,
    "StatusDescription": null,
    "Stable": null,
    "Version": null,
    "SubmitTime": null,
    "CreateIndex": null,
    "ModifyIndex": null,
    "JobModifyIndex": null
  }
""",
    )

    main(
        "Allocation",
        """
 {
    "ClientDescription": "Tasks are running",
    "ClientStatus": "running",
    "CreateIndex": 10,
    "CreateTime": 1636017249798459000,
    "DeploymentStatus": {
      "Canary": false,
      "Healthy": true,
      "ModifyIndex": 15,
      "Timestamp": "2021-11-04T10:14:22.054814+01:00"
    },
    "DesiredDescription": "",
    "DesiredStatus": "run",
    "DesiredTransition": {
      "ForceReschedule": null,
      "Migrate": null,
      "Reschedule": null
    },
    "EvalID": "cb20d15d-861f-8d8d-8253-e93932beea2e",
    "FollowupEvalID": "",
    "ID": "5457f16d-0f87-8e6b-5e91-0c7da3a41eb7",
    "JobID": "example",
    "JobType": "service",
    "JobVersion": 0,
    "ModifyIndex": 15,
    "ModifyTime": 1636017262190928000,
    "Name": "example.cache[0]",
    "Namespace": "default",
    "NodeID": "f476d2b4-02dc-c216-d031-273396727347",
    "NodeName": "linux",
    "PreemptedAllocations": null,
    "PreemptedByAllocation": "",
    "RescheduleTracker": null,
    "TaskGroup": "cache",
    "TaskStates": {
      "redis": {
        "Events": [
          {
            "Details": {},
            "DiskLimit": 0,
            "DisplayMessage": "Task received by client",
            "DownloadError": "",
            "DriverError": "",
            "DriverMessage": "",
            "ExitCode": 0,
            "FailedSibling": "",
            "FailsTask": false,
            "GenericSource": "",
            "KillError": "",
            "KillReason": "",
            "KillTimeout": 0,
            "Message": "",
            "RestartReason": "",
            "SetupError": "",
            "Signal": 0,
            "StartDelay": 0,
            "TaskSignal": "",
            "TaskSignalReason": "",
            "Time": 1636017249803624000,
            "Type": "Received",
            "ValidationError": "",
            "VaultError": ""
          },
          {
            "Details": {
              "message": "Building Task Directory"
            },
            "DiskLimit": 0,
            "DisplayMessage": "Building Task Directory",
            "DownloadError": "",
            "DriverError": "",
            "DriverMessage": "",
            "ExitCode": 0,
            "FailedSibling": "",
            "FailsTask": false,
            "GenericSource": "",
            "KillError": "",
            "KillReason": "",
            "KillTimeout": 0,
            "Message": "Building Task Directory",
            "RestartReason": "",
            "SetupError": "",
            "Signal": 0,
            "StartDelay": 0,
            "TaskSignal": "",
            "TaskSignalReason": "",
            "Time": 1636017249805254000,
            "Type": "Task Setup",
            "ValidationError": "",
            "VaultError": ""
          },
          {
            "Details": {},
            "DiskLimit": 0,
            "DisplayMessage": "Task started by client",
            "DownloadError": "",
            "DriverError": "",
            "DriverMessage": "",
            "ExitCode": 0,
            "FailedSibling": "",
            "FailsTask": false,
            "GenericSource": "",
            "KillError": "",
            "KillReason": "",
            "KillTimeout": 0,
            "Message": "",
            "RestartReason": "",
            "SetupError": "",
            "Signal": 0,
            "StartDelay": 0,
            "TaskSignal": "",
            "TaskSignalReason": "",
            "Time": 1636017252049956000,
            "Type": "Started",
            "ValidationError": "",
            "VaultError": ""
          }
        ],
        "Failed": false,
        "FinishedAt": null,
        "LastRestart": null,
        "Restarts": 0,
        "StartedAt": "2021-11-04T09:14:12.04996Z",
        "State": "running",
        "TaskHandle": null
      }
    }
  }
  """,
    )



#!/usr/bin/env python3

import argparse
import enum
import json
import logging
import re
import shlex
import subprocess
import sys
from abc import ABC
from dataclasses import dataclass
from pathlib import Path
from shlex import quote
from typing import List, Optional, Union

import click

from .common import common_options

log = logging.getLogger(Path(__file__).name)


def listquote(list: List[str]):
    return " ".join(quote(x) for x in list)


###############################################################################


@dataclass
class Mypath(ABC):
    path: Path

    def nomadrun(self, stdin: Union[bool, int] = False) -> str:
        return ""

    def isnomad(self):
        return False

    def exists(self):
        return self.path.exists()

    def is_file(self):
        return self.path.is_file()

    def is_dir(self):
        return self.path.is_dir()

    @property
    def name(self):
        return self.path.name

    def __str__(self):
        return str(self.path)

    def __truediv__(self, name: str):
        return type(self)(Path(f"{self}/{name}"))

    def quoteparent(self):
        return quote(str(self.path.parent))

    def quotename(self):
        return quote(str(self.path.name))

    def quotepath(self):
        return quote(str(self.path))


class FileType(enum.Enum):
    file = enum.auto()
    dir = enum.auto()
    none = enum.auto()


@dataclass
class NomadMypath(Mypath):
    nomadid: str
    _stat: Optional[FileType] = None

    def isnomad(self):
        return True

    def nomadrun(self, stdin: Union[bool, int] = False) -> str:
        # Return a properly escaped command line that allows to execute a command in the container.
        global args
        assert self.nomadid is not None
        idisjob = args.job or (
            # Try to be smart: if the id does not "look like" UUID number, assume it is a job name.
            not re.fullmatch("[0-9a-fA-F-]{1,36}", self.nomadid)
        )
        i = str(bool(stdin)).lower()
        return f"nomad alloc exec -t=false -i={i} " + (
            (
                (f"-namespace {quote(args.namespace)} " if args.namespace else "")
                + f"-job {quote(self.nomadid)}"
            )
            if idisjob
            else f"{quote(self.nomadid)}"
        )

    def _nomadstat(self) -> FileType:
        assert self.isnomad()
        if self._stat is None:
            # Execute a script inside the container to determine the type of the file.
            script = 'if [ -f "$@" ]; then echo file; elif [ -d "$@" ]; then echo dir; else echo none; fi'
            cmd = [
                *shlex.split(self.nomadrun()),
                "sh",
                "-c",
                script,
                "--",
                str(self.path),
            ]
            log.debug(f"+ {listquote(cmd)}")
            stat = subprocess.check_output(cmd, text=True).strip()
            log.debug(f"stat = {stat}")
            self._stat = FileType[stat]
        return self._stat

    def exists(self):
        return self._nomadstat() != FileType.none

    def is_file(self):
        return self._nomadstat() == FileType.file

    def is_dir(self):
        return self._nomadstat() == FileType.dir

    def __str__(self):
        return f"{self.nomadid}:{self.path}"


def mypath_factory(txt: str) -> Mypath:
    """
    ALLOCATION:SRC_PATH
    JOB:SRC_PATH
    JOB@TASK:SRC_PATH
    SRC_PATH
    -
    """
    arr = txt.split(":", 1)
    if len(arr) == 1:
        log.debug(f"Constructing Path({txt})")
        return Mypath(Path(txt))
    else:
        place = arr[0].split("@", 1)
        if len(place) == 1:
            log.debug(f"Constructing Nomad({arr[1]}, {arr[0]})")
            return NomadMypath(Path(arr[1]), arr[0])
        else:
            allocs = json.loads(
                subprocess.check_output(
                    "nomad operator api /v1/job/{place[0]}/allocations".split(),
                    text=True,
                )
            )
            allocs = [
                x
                for x in allocs
                if x["ClientStatus"] == "running" and place[1] in x["TaskStates"].keys()
            ]
            assert (
                len(allocs) < 2
            ), f"Multiple running allocations found for job {place[0]} and task {place[1]}: {' '.join(x['ID'] for x in allocs)}"
            assert (
                len(allocs) > 0
            ), f"No running allocations found for job {place[0]} and task {place[1]}"
            return NomadMypath(Path(arr[1]), allocs[0]["ID"])


###############################################################################


def bash(script: str):
    cmd = ["bash", "-o", "pipefail", "-c", script]
    log.debug(f"+ {listquote(cmd)}")
    subprocess.check_call(cmd)


def get_tar_x_opt():
    return (
        # If dry_run, pass vt to print to stdout.
        ("vt" if args.dry_run else "x")
        + ("v" if args.verbose else "")
        # If archive, pass p to preserve permissions when not root.
        # If not archive, pass o to do not preserve permissions even as root.
        # See tar docs.
        + ("p" if args.archive else "o")
    )


def tarpipe(src: Mypath, dst: Mypath, tar_c: str, tar_x: str):
    global args
    x = get_tar_x_opt()
    script = f"{src.nomadrun()} tar -cf - {tar_c} | {dst.nomadrun(stdin=1)} tar -{x}f - {tar_x}"
    bash(script)


def copy_mode(src: Mypath, dst: Mypath):
    # logic from https://docs.docker.com/engine/reference/commandline/cp/
    # SRC_PATH specifies a file
    if src.is_file():
        # DEST_PATH does not exist and ends with /
        if not dst.exists() and args.dest[-1:] == "/":
            exit("Error condition: the destination directory must exist.")
        # DEST_PATH is a directory
        if dst.is_dir():
            # the file is copied into this directory using the basename from SRC_PATH
            dst = dst / src.name
        if (
            # DEST_PATH does not exist
            # the file is saved to a file created at DEST_PATH
            not dst.exists()
            # DEST_PATH exists and is a file
            # the destination is overwritten with the source file’s contents
            or dst.is_file()
        ):
            # Tar the file.
            log.info(f"File {src} -> {dst}")
            # --transform the file to the output filename.
            tar_c = f"-C {src.quoteparent()} --transform s:.\\*:{dst.quotename()}: -- {src.quotename()}"
            # Unpack the file at destination filename.
            tar_x = f"-C {dst.quoteparent()} -- {dst.quotename()}"
            tarpipe(src, dst, tar_c, tar_x)
        else:
            exit(f"Not a file or directory: {dst}")
    # SRC_PATH specifies a directory
    elif src.is_dir():
        # DEST_PATH does not exist
        if not dst.exists():
            # DEST_PATH is created as a directory and the contents of the source directory are copied into this directory
            log.info(f"New dir {src} -> {dst}")
            # Use --transform to prepent all paths with destination directory name.
            tar_c = f"-C {src.quotepath()} --transform s:^:{dst.quotename()}/: ."
            # And unpack to parent directory.
            tar_x = f"-C {dst.quoteparent()}"
            tarpipe(src, dst, tar_c, tar_x)
        # DEST_PATH exists and is a file
        elif dst.is_file():
            exit("Error condition: cannot copy a directory to a file")
        # DEST_PATH exists and is a directory
        elif dst.is_dir():
            # SRC_PATH does not end with /. (that is: slash followed by dot)
            if args.source[-2:] != "/.":
                # the source directory is copied into this directory
                log.info(f"New dir {src} -> {dst}/{src.name}")
                tar_c = f"-C {src.quoteparent()} -- {src.quotename()}"
                tar_x = f"-C {dst.quotepath()}"
                tarpipe(src, dst, tar_c, tar_x)
            # SRC_PATH does end with /. (that is: slash followed by dot)
            else:
                # the content of the source directory is copied into this directory
                log.info(f"Content of {src}/. -> {dst}/.")
                tar_c = f"-C {src.quotepath()} ."
                tar_x = f"-C {dst.quotepath()}"
                tarpipe(src, dst, tar_c, tar_x)
        else:
            exit(f"Not a file or directory: {dst}")
    else:
        exit(f"Not a file or directory: {src}")


def stream_mode(src: Mypath, dst: Mypath):
    # One of source or dest is a -
    if args.source == "-":
        log.info(f"Stream stdin -> {args.dest}")
        assert args.dest != "-", "Both operands are '-'"
        x = get_tar_x_opt()
        bash(f"{dst.nomadrun(stdin=1)} tar -{x}f - -C {dst.quotepath()}")
    elif args.dest == "-":
        log.info(f"Stream {args.source} -> stdout")
        bash(f"{src.nomadrun()} tar -cf - -C {src.quoteparent()} -- {src.quotename()}")
    else:
        assert 0, "Internal error - neither source nor dest is equal to -"


def test():
    script = """
        set -xeuo pipefail
        cmd="$1"
        tmpd=$(mktemp -d)
        trap 'rm -vrf "$tmpd"' EXIT
        mkdir "$tmpd/src" "$tmpd/dst"
        echo 123 > "$tmpd/src/1"

        echo "testing file copy without destination"
        "$cmd" -d "$tmpd/src/1" "$tmpd/dst"
        test -e "$tmpd/dst/1"
        rm -v "$tmpd/dst/1"

        echo "tesing file copy to a different file"
        "$cmd" -d "$tmpd/src/1" "$tmpd/dst/2"
        test -e "$tmpd/dst/2"
        rm -v "$tmpd/dst/2"

        echo "testing dir copy to a dir"
        "$cmd" -d "$tmpd/src" "$tmpd/dst"
        test -e "$tmpd/dst/src/1"
        rm -vr "$tmpd/dst/src"

        echo "testing content of a dir copy to a dir"
        "$cmd" -d "$tmpd/src/." "$tmpd/dst"
        test -e "$tmpd/dst/1"
        rm -v "$tmpd/dst/1"
        """
    subprocess.check_call(
        [
            "bash",
            "-c",
            script,
            "--",
            sys.argv[0],
        ]
    )


###############################################################################


@click.command(
    help="""
Copy files/folders between a nomad allocation and the local filesystem.
Use '-' as the source to read a tar archive from stdin
and extract it to a directory destination in a container.
Use '-' as the destination to stream a tar archive of a
container source to stdout.

\b
Both source and dest take one of the forms:
    ALLOCATION:SRC_PATH
    JOB:SRC_PATH
    JOB@TASK:SRC_PATH
    SRC_PATH
    -

\b
Examples:
  {log.name} -n 9190d781:/tmp ~/tmp
  {log.name} -vn -Nservices -job promtail:/. ~/tmp
""",
    epilog=f"""
Written by Kamil Cukrowski 2023. Licensed under GNU GPL version or later.
""",
)
@click.option(
    "-n",
    "--dry-run",
    is_flag=True,
    help="Do tar -vt for unpacking. Usefull for listing files for debugging.",
)
@click.option("-v", "--verbose", count=True)
@click.option(
    "-N",
    "-namespace",
    "--namespace",
    help="Nomad namespace",
)
@click.option(
    "-a",
    "--archive",
    is_flag=True,
    help="Archive mode (copy all uid/gid information)",
)
@click.option(
    "-j",
    "-job",
    "--job",
    is_flag=True,
    help="Use a **random** allocation from the specified job ID.",
)
@click.option("--test", is_flag=True, help="Run tests")
@click.argument("source")
@click.argument("dest")
@common_options()
def cli(**kwargs):
    global args
    args = argparse.Namespace(**kwargs)
    logging.basicConfig(
        level=logging.DEBUG
        if args.verbose > 1
        else logging.INFO
        if args.verbose > 0
        else logging.WARNING,
        format="%(levelname)s %(name)s:%(funcName)s:%(lineno)d: %(message)s",
    )
    if args.test:
        test()
        exit()
    src = mypath_factory(args.source)
    dst = mypath_factory(args.dest)
    if args.source == "-" or args.dest == "-":
        stream_mode(src, dst)
    else:
        copy_mode(src, dst)


if __name__ == "__main__":
    cli()



from __future__ import annotations

import dataclasses
import datetime
import enum
import logging
import re
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar, Union

import dateutil.parser

from .datadict import DataDict

log = logging.getLogger(__name__)


class MyStrEnum(str, enum.Enum):
    """StrEnum for python 3.7 compatibility"""

    @staticmethod
    def _generate_next_value_(name, start, count, last_values):
        return name


class DockerMounts(DataDict):
    type: str
    target: str
    source: str
    readonly: bool


class JobTaskConfig(DataDict):
    image: str
    command: str
    args: List[str]
    network_mode: str
    network_aliases: List[str]
    volumes: List[str]
    mounts: List[DockerMounts]
    extra_hosts: List[str]


class LifecycleHook(MyStrEnum):
    prestart = "prestart"
    poststart = "poststart"
    poststop = "poststop"


class JobTaskLifecycle(DataDict):
    Hook: str
    Sidecar: bool

    def get_sidecar(self) -> bool:
        return self.get("Sidecar", False)


class JobTask(DataDict):
    Name: str
    Driver: str
    User: str
    Config: Union[dict, JobTaskConfig]
    Lifecycle: Optional[JobTaskLifecycle] = None
    Env: Optional[Dict[str, str]]
    Services: Optional[List[Any]]


class JobTaskGroup(DataDict):
    Name: str
    Count: int
    Tasks: List[JobTask]
    Services: Optional[List[Any]]


class JobStatus(MyStrEnum):
    pending = enum.auto()
    """Pending means the job is waiting on scheduling"""
    running = enum.auto()
    """Running means the job has non-terminal allocations"""
    dead = enum.auto()
    """Dead means all evaluation's and allocations are terminal"""


class Job(DataDict):
    """Returned aby job/<id> API. DO NOT mix with JobsJob"""

    ID: str
    Version: int
    Status: str
    Namespace: str
    ModifyIndex: int
    JobModifyIndex: int
    TaskGroups: List[JobTaskGroup]
    Stop: bool
    Meta: Optional[Dict[str, str]]
    SubmitTime: int

    def is_dead(self):
        return self.Status == JobStatus.dead

    def description(self):
        return f"{self.ID}#{self.Version}@{self.Namespace}"

    def allservices(self):
        """Return all services from groups and tasks of a job"""
        out = []
        for tg in self.TaskGroups:
            if tg.Services:
                out += tg.Services
            for task in tg.Tasks:
                if task.Services:
                    out += task.Services
        return out


class JobsJob(DataDict):
    """Returned aby jobs API. DO NOT mix with Job"""

    ID: str


class NodeScoreMeta(DataDict):
    NodeID: str
    Scores: Dict[str, float]
    NormScore: float


class AllocationMetric(DataDict):
    NodesEvaluated: int
    NodesFiltered: int
    NodesAvailable: Dict[str, int]
    ClassFiltered: Optional[Dict[str, int]] = None
    ConstraintFiltered: Optional[Dict[str, int]] = None
    NodesExhausted: int
    QuotaExhausted: Optional[Dict[str, int]] = None
    ScoreMetaData: Optional[List[NodeScoreMeta]] = None
    ClassExhausted: Optional[Dict[str, int]] = None
    Scores: Optional[Dict[str, float]] = None
    DimensionExhausted: Optional[Dict[str, int]] = None

    def format(self, scores: bool = False, prefix: str = "") -> str:
        """https://github.com/hashicorp/nomad/blob/484f91b893c6f054e9339f8db6bd157429c2ef20/command/monitor.go#L345"""
        out = []
        if self.NodesEvaluated == 0:
            out += [f"No nodes were eligible for evaluation"]
        for dc, available in self.NodesAvailable.items():
            if available == 0:
                out += [f"No nodes are available in datacenter {dc}"]
        for cls, num in (self.ClassFiltered or {}).items():
            out += [f"Class {cls}: {num} nodes excluded by filter"]
        for cs, num in (self.ConstraintFiltered or {}).items():
            out += [f"Constraint {cs}: {num} nodes excluded by filter"]
        ne = self.NodesExhausted
        if ne > 0:
            out += [f"Resources exhaused on {ne} nodes"]
        for cls, num in (self.ClassExhausted or {}).items():
            out += [f"Class {cls} exhaused on {num} nodes"]
        for dim, num in (self.DimensionExhausted or {}).items():
            out += [f"Dimension {dim} exhaused on {num} nodes"]
        for _, dim in (self.QuotaExhausted or {}).items():
            out += [f"Quota limit hit {dim}"]
        if scores:
            if self.ScoreMetaData:
                allScores: List[str] = sorted(
                    list(set(k for s in self.ScoreMetaData for k in s.Scores.keys()))
                )
                out += [f"Node|{'|'.join(allScores)}|final score"]
                for scoreMeta in self.ScoreMetaData:
                    out += [
                        f"{scoreMeta.NodeID}|"
                        + "|".join(
                            f"{scoreMeta.Scores[scorerName]:.3g}"
                            for scorerName in allScores
                        )
                        + f"|{scoreMeta.NormScore:.3g}"
                    ]
            else:
                for name, score in (self.Scores or {}).items():
                    out += [f"Score {name} = {score}"]
        return "\n".join(f"{prefix}{x}" for x in out)


class EvalStatus(MyStrEnum):
    blocked = enum.auto()
    pending = enum.auto()
    complete = enum.auto()
    failed = enum.auto()
    canceled = enum.auto()


def fromisoformat(txt: str) -> datetime.datetime:
    return dateutil.parser.isoparse(txt)


class Eval(DataDict):
    ID: str
    Namespace: str
    JobID: str
    JobModifyIndex: int
    ModifyIndex: int
    ModifyTime: int
    Status: str
    WaitUntil: str
    FailedTGAllocs: Dict[str, AllocationMetric]

    def is_pending(self):
        return self.Status == "pending"

    def getWaitUntil(self) -> Optional[datetime.datetime]:
        if not self.WaitUntil:
            return None
        return (
            fromisoformat(self.WaitUntil)
            .replace(tzinfo=datetime.timezone.utc)
            .astimezone()
        )


class AllocTaskStateEventType(MyStrEnum):
    TaskSetupFailure = "Setup Failure"
    """indicates that the task could not be started due to a a setup failure."""
    TaskDriverFailure = "Driver Failure"
    """indicates that the task could not be started due to a failure in the driver. TaskDriverFailure is considered Recoverable."""
    TaskReceived = "Received"
    """signals that the task has been pulled by the client at the given timestamp."""
    TaskFailedValidation = "Failed Validation"
    """indicates the task was invalid and as such was not run. TaskFailedValidation is not considered Recoverable."""
    TaskStarted = "Started"
    """signals that the task was started and its timestamp can be used to determine the running length of the task."""
    TaskTerminated = "Terminated"
    """indicates that the task was started and exited."""
    TaskKilling = "Killing"
    """indicates a kill signal has been sent to the task."""
    TaskKilled = "Killed"
    """indicates a user has killed the task."""
    TaskRestarting = "Restarting"
    """indicates that task terminated and is being restarted."""
    TaskNotRestarting = "Not Restarting"
    """indicates that the task has failed and is not being restarted because it has exceeded its restart policy."""
    TaskRestartSignal = "Restart Signaled"
    """indicates that the task has been signaled to be restarted"""
    TaskSignaling = "Signaling"
    """indicates that the task is being signalled."""
    TaskDownloadingArtifacts = "Downloading Artifacts"
    """means the task is downloading the artifacts specified in the task."""
    TaskArtifactDownloadFailed = "Failed Artifact Download"
    """indicates that downloading the artifacts failed."""
    TaskBuildingTaskDir = "Building Task Directory"
    """indicates that the task directory/chroot is being built."""
    TaskSetup = "Task Setup"
    """indicates the task runner is setting up the task environment"""
    TaskDiskExceeded = "Disk Resources Exceeded"
    """indicates that one of the tasks in a taskgroup has exceeded the requested disk resources."""
    TaskSiblingFailed = "Sibling Task Failed"
    """indicates that a sibling task in the task group has failed."""
    TaskDriverMessage = "Driver"
    """is an informational event message emitted by drivers such as when they're performing a long running action like downloading an image."""
    TaskLeaderDead = "Leader Task Dead"
    """indicates that the leader task within the has finished."""
    TaskMainDead = "Main Tasks Dead"
    """indicates that the main tasks have dead"""
    TaskHookFailed = "Task hook failed"
    """indicates that one of the hooks for a task failed."""
    TaskHookMessage = "Task hook message"
    """indicates that one of the hooks for a task emitted a message."""
    TaskRestoreFailed = "Failed Restoring Task"
    """indicates Nomad was unable to reattach to a restored task."""
    TaskPluginUnhealthy = "Plugin became unhealthy"
    """indicates that a plugin managed by Nomad became unhealthy"""
    TaskPluginHealthy = "Plugin became healthy"
    """indicates that a plugin managed by Nomad became healthy"""
    TaskClientReconnected = "Reconnected"
    """indicates that the client running the task disconnected."""
    TaskWaitingShuttingDownDelay = "Waiting for shutdown delay"
    """indicates that the task is waiting for shutdown delay before being TaskKilled"""
    TaskSkippingShutdownDelay = "Skipping shutdown delay"
    """indicates that the task operation was configured to ignore the shutdown delay value set for the tas."""


class AllocTaskStateEvent(DataDict):
    DisplayMessage: str
    Time: int
    Type: str


class AllocTaskState(DataDict):
    Events: Optional[List[AllocTaskStateEvent]]
    State: str
    Failed: bool
    FinishedAt: Optional[str]
    LastRestart: Optional[str]
    Restarts: int
    StartedAt: Optional[str]

    def find_event(self, type_: str) -> Optional[AllocTaskStateEvent]:
        """Find event in TaskStates task Events. Return empty dict if not found"""
        return next((e for e in self.Events or [] if e.Type == type_), None)

    def was_started(self):
        return self.find_event(AllocTaskStateEventType.TaskStarted) is not None


class Alloc(DataDict):
    ID: str
    NodeName: str
    JobID: str
    EvalID: str
    ClientStatus: str
    CreateTime: int
    Namespace: str
    ModifyIndex: int
    ModifyTime: int
    TaskGroup: str
    # Also may be missing!
    TaskStates: Optional[Dict[str, AllocTaskState]] = None
    # May be missing!
    JobVersion: Optional[int]
    FollowupEvalID: Optional[str] = None

    def get_taskstates(self) -> Dict[str, AllocTaskState]:
        """The same as TaskStates but returns an empty dict in case the field is None"""
        return self.get("TaskStates") or {}

    def get_tasknames(self):
        return [k for k in self.get_taskstates().keys()]

    def is_pending_or_running(self):
        return self.ClientStatus in ["pending", "running"]

    def is_running(self):
        return self.ClientStatus == "running"

    def is_finished(self):
        return not self.is_pending_or_running()


class DeploymentStatus(MyStrEnum):
    running = enum.auto()
    paused = enum.auto()
    failed = enum.auto()
    successful = enum.auto()
    cancelled = enum.auto()
    initializing = enum.auto()
    pending = enum.auto()
    blocked = enum.auto()
    unblocking = enum.auto()


class DeploymentStatusDescription(MyStrEnum):
    Running = "Deployment is running"
    RunningNeedsPromotion = "Deployment is running but requires manual promotion"
    RunningAutoPromotion = "Deployment is running pending automatic promotion"
    Paused = "Deployment is paused"
    Successful = "Deployment completed successfully"
    StoppedJob = "Cancelled because job is stopped"
    NewerJob = "Cancelled due to newer version of job"
    FailedAllocations = "Failed due to unhealthy allocations"
    ProgressDeadline = "Failed due to progress deadline"
    FailedByUser = "Deployment marked as failed"
    FailedByPeer = "Failed because of an error in peer region"
    Blocked = "Deployment is complete but waiting for peer region"
    Unblocking = "Deployment is unblocking remaining regions"
    PendingForPeer = "Deployment is pending, waiting for peer region"


class DeploymentTaskGroup(DataDict):
    AutoPromote: bool
    AutoRevert: bool
    DesiredCanaries: int
    DesiredTotal: int
    HealthyAllocs: int
    PlacedAllocs: int
    PlacedCanaries: Optional[List[str]] = None
    ProgressDeadline: int
    Promoted: bool
    RequireProgressBy: Optional[str]
    UnhealthyAllocs: int


class Deploy(DataDict):
    ID: str
    ModifyIndex: int
    JobCreateIndex: int
    JobModifyIndex: int
    JobVersion: int
    JobID: str
    Status: str
    StatusDescription: str
    TaskGroups: Dict[str, DeploymentTaskGroup]


class EventTopic(enum.Enum):
    """Topic of an event from Nomad event stream"""

    ACLToken = enum.auto()
    ACLPolicy = enum.auto()
    ACLRoles = enum.auto()
    Allocation = enum.auto()
    Job = enum.auto()
    Evaluation = enum.auto()
    Deployment = enum.auto()
    Node = enum.auto()
    NodeDrain = enum.auto()
    NodePool = enum.auto()
    Service = enum.auto()


class EventType(MyStrEnum):
    """Type of an event from Nomad event stream"""

    ACLTokenUpserted = enum.auto()
    ACLTokenDeleted = enum.auto()
    ACLPolicyUpserted = enum.auto()
    ACLPolicyDeleted = enum.auto()
    ACLRoleUpserted = enum.auto()
    ACLRoleDeleted = enum.auto()
    AllocationCreated = enum.auto()
    AllocationUpdated = enum.auto()
    AllocationUpdateDesiredStatus = enum.auto()
    DeploymentStatusUpdate = enum.auto()
    DeploymentPromotion = enum.auto()
    DeploymentAllocHealth = enum.auto()
    EvaluationUpdated = enum.auto()
    JobRegistered = enum.auto()
    JobDeregistered = enum.auto()
    JobBatchDeregistered = enum.auto()
    NodeRegistration = enum.auto()
    NodeDeregistration = enum.auto()
    NodeEligibility = enum.auto()
    NodeDrain = enum.auto()
    NodeEvent = enum.auto()
    NodePoolUpserted = enum.auto()
    NodePoolDeleted = enum.auto()
    PlanResult = enum.auto()
    ServiceRegistration = enum.auto()
    ServiceDeregistration = enum.auto()


R = TypeVar("R")


@dataclasses.dataclass
class Event:
    """A single Event as returned from Nomad event stream"""

    topic: EventTopic
    type: EventType
    data: dict
    time: Optional[datetime.datetime] = None
    stream: bool = False

    def __str__(self):
        status = {
            "ID": self.data.get("ID", "")[:6],
            "JobID": self.data.get("JobID"),
            "Version": self.data.get("Version"),
            "JobVersion": self.data.get("JobVersion"),
            "JobModifyIndex": self.data.get("JobModifyIndex"),
            "ModifyIndex": self.data.get("ModifyIndex"),
            "Status": self.data.get("Status"),
            "ClientStatus": self.data.get("ClientStatus"),
            "stream": 1 if self.stream else 0,
        }
        statusstr = " ".join(f"{k}={v}" for k, v in status.items() if v)
        return f"Event({self.topic.name}.{self.type.name} {statusstr})"

    def is_job(self):
        return self.topic == EventTopic.Job

    def is_eval(self):
        return self.topic == EventTopic.Evaluation

    def is_alloc(self):
        return self.topic == EventTopic.Allocation

    def is_deployment(self):
        return self.topic == EventTopic.Deployment

    def get_job(self) -> Optional[Job]:
        return Job(self.data) if self.is_job() else None

    def get_eval(self) -> Optional[Eval]:
        return Eval(self.data) if self.is_eval() else None

    def get_alloc(self) -> Optional[Alloc]:
        return Alloc(self.data) if self.is_alloc() else None

    def get_deployment(self) -> Optional[Deploy]:
        return Deploy(self.data) if self.is_deployment() else None

    def apply(
        self,
        job: Optional[Callable[[Job], R]] = None,
        eval: Optional[Callable[[Eval], R]] = None,
        alloc: Optional[Callable[[Alloc], R]] = None,
        deploy: Optional[Callable[[Deploy], R]] = None,
    ) -> R:
        callbacks = {}
        if job:
            callbacks[EventTopic.Job] = lambda data: job(Job(data))
        if eval:
            callbacks[EventTopic.Evaluation] = lambda data: eval(Eval(data))
        if alloc:
            callbacks[EventTopic.Allocation] = lambda data: alloc(Alloc(data))
        if deploy:
            callbacks[EventTopic.Deployment] = lambda data: deploy(Deploy(data))
        assert len(callbacks), f"At least one callback has to be specified"
        return callbacks[self.topic](self.data)


class EventApplier(ABC, Generic[R]):
    @abstractmethod
    def apply_job(self, job: Job) -> R:
        raise NotImplementedError()

    @abstractmethod
    def apply_eval(self, eval: Eval) -> R:
        raise NotImplementedError()

    @abstractmethod
    def apply_alloc(self, alloc: Alloc) -> R:
        raise NotImplementedError()

    @abstractmethod
    def apply_deploy(self, deploy: Deploy) -> R:
        raise NotImplementedError()

    def apply(self, e: Event) -> R:
        if e.is_job():
            return self.apply_job(Job(e.data))
        elif e.is_alloc():
            return self.apply_alloc(Alloc(e.data))
        elif e.is_eval():
            return self.apply_eval(Eval(e.data))
        elif e.is_deployment():
            return self.apply_deploy(Deploy(e.data))
        else:
            raise KeyError(f"{e.topic} not handled")


class VariableNoItems(DataDict):
    Namespace: str
    Path: str
    CreateIndex: int
    ModifyIndex: int
    CreateTime: int
    ModifyTime: int


class Variable(VariableNoItems):
    Items: Dict[str, str]


class VariableNew(DataDict):
    Namespace: str
    Path: str
    Items: Dict[str, str]


class JobSummaryChildren(DataDict):
    Pending: int
    Running: int
    Dead: int


class JobSummarySummary(DataDict):
    Queued: int = 0
    Complete: int = 0
    Failed: int = 0
    Running: int = 0
    Starting: int = 0
    Lost: int = 0

    def __iadd__(self, o: JobSummarySummary) -> JobSummarySummary:
        for k in set(self.asdict()) | set(o.asdict()):
            self[k] = self.get(k, 0) + o.get(k, 0)
        return self


class JobSummary(DataDict):
    JobID: str
    Summary: Dict[str, JobSummarySummary]
    Children: JobSummaryChildren
    CreateIndex: int
    ModifyIndex: int

    def get_sum_summary(self) -> JobSummarySummary:
        """Sum all summaries into one"""
        ret = JobSummarySummary({})
        for s in self.Summary.values():
            ret += s
        return ret



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import argparse
import base64
import dataclasses
import datetime
import enum
import inspect
import itertools
import json
import logging
import re
import shlex
import subprocess
import threading
import time
from abc import ABC, abstractmethod
from http import client as http_client
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Pattern,
    Set,
    Tuple,
    TypeVar,
)

import click
import requests

from . import colors, exit_on_thread_exception, nomadlib
from .common import (
    _complete_set_namespace,
    alias_option,
    common_options,
    complete_job,
    completor,
    composed,
    json_loads,
    mynomad,
    namespace_option,
    nomad_find_job,
)
from .nomaddbjob import NomadDbJob
from .nomadlib import Event, EventTopic, EventType, MyStrEnum, ns2dt

log = logging.getLogger(__name__)

###############################################################################

args = argparse.Namespace()
"""Arguments passed to this program"""

START_NS: int = 0
"""This program start time, assigned from cli()"""

COLORS = colors.init()

T = TypeVar("T")


def set_not_in_add(s: Set[T], value: T) -> bool:
    """If value is in the set, return False. Otherwise add the value to the set and return True"""
    if value in s:
        return False
    s.add(value)
    return True


def andjoin(arr: Iterable[str]) -> str:
    arr = list(arr)
    if not len(arr):
        return ""
    if len(arr) == 1:
        return arr[0]
    return ", ".join(arr[:-1]) + " and " + arr[-1]


###############################################################################


class LogWhat(MyStrEnum):
    deploy = enum.auto()
    eval = enum.auto()
    alloc = enum.auto()
    stdout = enum.auto()
    stderr = enum.auto()


class LOGENABLED:
    """Should logging of a specific stream be enabled?"""

    deploy: bool = True
    eval: bool = True
    alloc: bool = True
    stdout: bool = True
    stderr: bool = True


class LOGFORMAT:
    """Logging output format specification templates using f-string. Very poor mans templating langauge."""

    pre = "{color}{now.strftime(args.log_time_format) + '>' if args.log_time else ''}"
    mark = "{mark}>"
    post = "{message}{reset}"
    task = "{task + '>' if task else ''}"
    DEFAULT = (
        pre + mark + "{id:.{args.log_id_len}}>#{str(jobversion)}>" + task + " " + post
    )
    """
    Default log format. The log is templated with f-string using eval() below.
        O>45fbbd>#0>group1>task1> hello world
    """
    ONE = pre + mark + task + " " + post
    """
    Log format with -1 option.
        O>task1> hello world
    """
    ZERO = pre + post
    """
    Log format with -0 option.
        hello world
    """
    LOGGING = (
        COLORS.blue
        + "{'%(asctime)s>' if args.log_time else ''}"
        + "{'nomad-watch' if args.verbose <= 0 else '%(module)s'}>"
        + "%(lineno)03d>"
        + " %(levelname)s %(message)s"
        + COLORS.reset
    )
    """Logging format, first templated with f-string, then by logging."""
    colors: Dict[LogWhat, str] = {
        LogWhat.deploy: COLORS.brightmagenta,
        LogWhat.eval: COLORS.magenta,
        LogWhat.alloc: COLORS.cyan,
        LogWhat.stderr: COLORS.orange,
        LogWhat.stdout: "",
    }
    """Logs colors"""
    marks: Dict[LogWhat, str] = {
        LogWhat.deploy: "deploy",
        LogWhat.eval: "eval",
        LogWhat.alloc: "A",
        LogWhat.stderr: "E",
        LogWhat.stdout: "O",
    }
    """Logs marks"""

    @staticmethod
    def render(fmt: str, **kwargs) -> str:
        locals().update(kwargs)
        return eval(f"f{fmt!r}")


def click_log_options():
    """All logging options"""
    return composed(
        click.option(
            "-T",
            "--log-time",
            is_flag=True,
            help="Additionally add timestamp to the logs. The timestamp of stdout and stderr streams is when the log was received, as Nomad does not store timestamp of task logs.",
        ),
        click.option(
            "--log-time-format",
            default="%Y-%m-%dT%H:%M:%S%z",
            show_default=True,
            help="Format time with specific format. Passed to python datetime.strftime.",
        ),
        alias_option(
            "-H",
            "--log-time-hour",
            aliased=dict(
                log_time_format="%H:%M:%S",
                log_time=True,
            ),
        ),
        click.option(
            "--log-format",
            default=LOGFORMAT.DEFAULT,
            show_default=True,
            help="The format to use when printing job logs",
        ),
        click.option(
            "--log-id-len",
            default=6,
            help="The length of id to log. UUIDv4 has 36 characters.",
        ),
        alias_option(
            "-l",
            "--log-id-long",
            aliased=dict(log_id_len=36),
        ),
        alias_option(
            "-1",
            "--log-only-task",
            aliased=dict(log_format=LOGFORMAT.ONE),
        ),
        alias_option(
            "-0",
            "--log-none",
            aliased=dict(log_format=LOGFORMAT.ZERO),
        ),
    )


def init_logging():
    LOGENABLED.deploy = any(s in "all deployment deploy d".split() for s in args.out)
    LOGENABLED.eval = any(s in "all evaluation eval e".split() for s in args.out)
    LOGENABLED.alloc = any(s in "all alloc a".split() for s in args.out)
    LOGENABLED.stderr = any(s in "all stderr err E 2".split() for s in args.out)
    LOGENABLED.stdout = any(s in "all stdout out O 1".split() for s in args.out)
    #
    args.verbose -= args.quiet
    if args.verbose > 1:
        http_client.HTTPConnection.debuglevel = 1
    logging.basicConfig(
        format=LOGFORMAT.render(LOGFORMAT.LOGGING),
        datefmt=args.log_time_format,
        level=(
            logging.DEBUG
            if args.verbose > 0
            else logging.INFO
            if args.verbose == 0
            else logging.WARN
            if args.verbose == 1
            else logging.ERROR
        ),
    )


@dataclasses.dataclass(frozen=True)
class Mylogger:
    """Used to log from the various streams we want to log from"""

    id: str
    message: str
    what: str
    now: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)
    nodename: Optional[str] = None
    jobversion: Optional[str] = None
    group: Optional[str] = None
    task: Optional[str] = None

    def fmt(self):
        assert self.what in [
            e.value for e in LogWhat
        ], f"{self.what} value is not a valid what"
        mark: str = LOGFORMAT.marks[self.what]
        color: str = LOGFORMAT.colors[self.what]
        reset: str = COLORS.reset if color else ""
        return LOGFORMAT.render(
            args.log_format,
            **dataclasses.asdict(self),
            mark=mark,
            color=color,
            reset=reset,
            args=args,
        )

    @staticmethod
    def __log(what: str, **kwargs):
        mylogger = Mylogger(what=what, **kwargs)
        try:
            out = mylogger.fmt()
        except KeyError:
            log.exception(f"fmt={args.log_format!r} params={args}")
            raise
        print(out, flush=True)

    @classmethod
    def log_eval(cls, db: NomadDbJob, eval: nomadlib.Eval, message: str):
        return cls.__log(
            LogWhat.eval,
            id=eval.ID,
            jobversion=db.find_jobversion_from_modifyindex(eval.ModifyIndex),
            now=ns2dt(eval.ModifyTime),
            message=message,
        )

    @classmethod
    def log_deploy(cls, db: NomadDbJob, deploy: nomadlib.Deploy, message: str):
        job = db.jobversions.get(deploy.JobVersion)
        return cls.__log(
            LogWhat.deploy,
            id=deploy.ID,
            jobversion=deploy.JobVersion,
            now=ns2dt(job.SubmitTime) if job else datetime.datetime.now(),
            message=message,
        )

    @classmethod
    def log_alloc(cls, allocid: str, **kwargs):
        return cls.__log(LogWhat.alloc, id=allocid, **kwargs)

    @classmethod
    def log_std(cls, stderr: bool, allocid: str, **kwargs):
        return cls.__log(
            LogWhat.stderr if stderr else LogWhat.stdout, id=allocid, **kwargs
        )


@dataclasses.dataclass(frozen=True)
class TaskKey:
    """Represent data to unique identify a task"""

    allocid: str
    jobversion: str
    nodename: str
    group: str
    task: str

    def __str__(self):
        return f"{self.allocid:.6}:v{self.jobversion}:{self.group}:{self.task}"

    def asdict(self):
        return dataclasses.asdict(self)

    def log_alloc(self, now: datetime.datetime, message: str):
        Mylogger.log_alloc(now=now, message=message, **self.asdict())

    def log_task(self, stderr: bool, message: str):
        Mylogger.log_std(
            stderr=stderr,
            message=message,
            now=datetime.datetime.now().astimezone(),
            **self.asdict(),
        )


###############################################################################


class TaskLogger(threading.Thread):
    """Represents a single logging stream from Nomad. Such stream is created separately for stdout and stderr."""

    def __init__(self, tk: TaskKey, stderr: bool):
        super().__init__(name=f"{tk}{1 + int(stderr)}")
        self.tk = tk
        self.stderr: bool = stderr
        self.exitevent = threading.Event()
        self.ignoredlines: List[str] = []
        self.first_line = True
        # Ignore input lines if printing only trailing lines.
        self.ignoretime_ns = START_NS + int(args.lines_timeout * 1e9)
        # If ignore time is in the past, it is no longer relevant anyway.
        if args.lines < 0 or self.ignoretime_ns < time.time_ns():
            self.ignoretime_ns = 0

    @staticmethod
    def read_json_stream(stream: requests.Response):
        txt: str = ""
        for data in stream.iter_content(decode_unicode=True):
            for c in data:
                txt += c
                # Nomad happens to be consistent, the jsons are flat.
                if c == "}":
                    try:
                        ret = json_loads(txt)
                        # log.debug(f"RECV: {ret}")
                        yield ret
                    except json.JSONDecodeError as e:
                        log.warn(f"error decoding json: {txt} {e}")
                    txt = ""

    def taskout(self, lines: List[str]):
        """Output the lines"""
        # If ignoring and this is first received line or the ignoring time is still happenning.
        if self.ignoretime_ns and (
            self.first_line or time.time_ns() < self.ignoretime_ns
        ):
            # Accumulate args.lines into ignoredlines array.
            self.first_line = False
            self.ignoredlines = lines
            self.ignoredlines = self.ignoredlines[: args.lines]
        else:
            if self.ignoretime_ns:
                # If not ignoring lines, flush the accumulated lines.
                lines = self.ignoredlines + lines
                self.ignoredlines.clear()
                # Disable further accumulation of ignored lines.
                self.ignoretime_ns = 0
            # Print the log lines.
            for line in lines:
                line = line.rstrip()
                self.tk.log_task(self.stderr, line)

    def __typestr(self):
        return "stderr" if self.stderr else "stdout"

    def __run_in(self):
        with mynomad.stream(
            f"client/fs/logs/{self.tk.allocid}",
            params={
                "task": self.tk.task,
                "type": self.__typestr(),
                "follow": True,
                "origin": "end" if self.ignoretime_ns else "start",
                "offset": 50000 if self.ignoretime_ns else 0,
            },
        ) as stream:
            for event in self.read_json_stream(stream):
                if event:
                    line64: Optional[str] = event.get("Data")
                    if line64:
                        lines = base64.b64decode(line64.encode()).decode().splitlines()
                        self.taskout(lines)
                else:
                    # Nomad json stream periodically sends empty {}.
                    # No idea why, but I can implement timeout.
                    self.taskout([])
                    if self.exitevent.is_set():
                        break

    def run(self):
        """Listen to Nomad log stream and print the logs"""
        try:
            self.__run_in()
        except requests.HTTPError as e:
            if e.response and e.response.status_code == 404:
                self.tk.log_alloc(
                    datetime.datetime.now(),
                    f"{self.__typestr()} logs were garbage collected from Nomad",
                )
            else:
                raise

    def stop(self):
        self.exitevent.set()


@dataclasses.dataclass
class TaskHandler:
    """A handler for one task. Creates loggers, writes out task events, handle exit conditions"""

    loggers: Optional[List[TaskLogger]] = None
    """Array of loggers that log allocation logs."""
    messages: Set[int] = dataclasses.field(default_factory=set)
    """A set of message timestamp to know what has been printed."""
    exitcode: Optional[int] = None
    stoptimer: Optional[threading.Timer] = None

    @staticmethod
    def _create_loggers(tk: TaskKey):
        ths: List[TaskLogger] = []
        if LOGENABLED.stdout:
            ths.append(TaskLogger(tk, False))
        if LOGENABLED.stderr:
            ths.append(TaskLogger(tk, True))
        for th in ths:
            th.start()
        log.debug(f"Started {len(ths)} loggers for {tk}")
        return ths

    def notify(self, tk: TaskKey, taskstate: nomadlib.AllocTaskState):
        """Receive notification that a task state has changed"""
        if LOGENABLED.alloc:
            for e in taskstate.Events or []:
                msgtime_ns = e.Time
                # Ignore message before ignore times.
                if (
                    msgtime_ns
                    and (
                        args.lines < 0
                        or msgtime_ns >= START_NS
                        or len(self.messages) < args.lines
                    )
                    and set_not_in_add(self.messages, msgtime_ns)
                ):
                    for line in e.DisplayMessage.splitlines():
                        if line:
                            tk.log_alloc(ns2dt(msgtime_ns), f"{e.Type} {line}")
        if (
            self.loggers is None
            and taskstate.State in ["running", "dead"]
            and taskstate.was_started()
        ):
            self.loggers = self._create_loggers(tk)
        if self.stoptimer is None and self.loggers and taskstate.State == "dead":
            # If the task is already finished, give myself max 3 seconds to query all the logs.
            # This is to reduce the number of connections.
            self.stoptimer = threading.Timer(3, self.stop)
            self.stoptimer.start()
        if self.exitcode is None and taskstate.State == "dead":
            terminatedevent = taskstate.find_event("Terminated")
            if terminatedevent:
                self.exitcode = terminatedevent["ExitCode"]

    def stop(self):
        if self.stoptimer:
            self.stoptimer.cancel()
        for l in self.loggers or []:
            l.stop()


@dataclasses.dataclass
class AllocWorker:
    """Represents a worker that prints out and manages state related to one allocation"""

    db: NomadDbJob
    taskhandlers: Dict[TaskKey, TaskHandler] = dataclasses.field(default_factory=dict)

    def notify(self, alloc: nomadlib.Alloc):
        """Update the state with alloc"""
        for taskname, task in alloc.get_taskstates().items():
            if args.task and not args.task.search(taskname):
                continue
            tk = TaskKey(
                alloc.ID,
                str(self.db.get_allocation_job_version(alloc, "?")),
                alloc.NodeName,
                alloc.TaskGroup,
                taskname,
            )
            self.taskhandlers.setdefault(tk, TaskHandler()).notify(tk, task)


class ExitCode:
    success = 0
    exception = 1
    interrupted = 2
    """This program execution flow was interrupted - user clicked ctrl+c"""
    failed = 3
    """Starting a job failed - deployment was reverted"""
    any_failed_tasks = 124
    all_failed_tasks = 125
    any_unfinished_tasks = 126
    no_allocations = 127


@dataclasses.dataclass
class NotifierWorker:
    """An containers for storing a map of allocation workers"""

    db: NomadDbJob
    """Link to database"""
    workers: Dict[str, AllocWorker] = dataclasses.field(default_factory=dict)
    messages: Set[Tuple[int, str]] = dataclasses.field(default_factory=set)
    """A set of evaluation ModifyIndex to know what has been printed."""

    def lineno_key_not_printed(self, key: str) -> bool:
        lineno = inspect.currentframe().f_back.f_lineno  # type: ignore
        return set_not_in_add(self.messages, (lineno, key))

    def notify_alloc(self, alloc: nomadlib.Alloc):
        if LOGENABLED.eval:
            evaluation = self.db.evaluations.get(alloc.EvalID)
            if evaluation and self.lineno_key_not_printed(alloc.EvalID):
                Mylogger.log_eval(
                    self.db,
                    evaluation,
                    f"Allocation {alloc.ID} started on {alloc.NodeName}",
                )
        #
        self.workers.setdefault(alloc.ID, AllocWorker(self.db)).notify(alloc)
        #
        if LOGENABLED.eval and alloc.is_finished():
            evaluation = self.db.evaluations.get(alloc.EvalID)
            if evaluation and self.lineno_key_not_printed(alloc.EvalID):
                Mylogger.log_eval(
                    self.db, evaluation, f"Allocation {alloc.ID} finished"
                )
        #
        if LOGENABLED.eval and alloc.FollowupEvalID:
            followupeval = self.db.evaluations.get(alloc.FollowupEvalID)
            if followupeval:
                waituntil = followupeval.getWaitUntil()
                if waituntil and self.lineno_key_not_printed(followupeval.ID):
                    utcnow = datetime.datetime.now(datetime.timezone.utc)
                    delay = waituntil - utcnow
                    if delay > datetime.timedelta(0):
                        Mylogger.log_eval(
                            self.db,
                            followupeval,
                            f"Nomad will attempt to reschedule in {delay} seconds",
                        )

    def notify_eval(self, evaluation: nomadlib.Eval):
        if (
            LOGENABLED.eval
            and evaluation.Status == nomadlib.EvalStatus.blocked
            and "FailedTGAllocs" in evaluation
            and self.lineno_key_not_printed(evaluation.ID)
        ):
            Mylogger.log_eval(
                self.db,
                evaluation,
                f"{evaluation.JobID}: Placement Failures: {len(evaluation.FailedTGAllocs)} unplaced",
            )
            for task, metric in evaluation.FailedTGAllocs.items():
                for msg in metric.format(True, f"{task}: ").splitlines():
                    Mylogger.log_eval(self.db, evaluation, msg)

    def notify_deploy(self, deployment: nomadlib.Deploy):
        # If the job has any service defined.
        if (
            LOGENABLED.eval
            and self.db.job
            and deployment.Status
            in [
                nomadlib.DeploymentStatus.successful,
                nomadlib.DeploymentStatus.failed,
            ]
        ):
            """https://github.com/hashicorp/nomad/blob/e02dd2a331c778399fd271e85c75bff3e3783d80/ui/app/templates/components/job-deployment/deployment-metrics.hbs#L10"""
            for task, tg in deployment.TaskGroups.items():
                Mylogger.log_deploy(
                    self.db,
                    deployment,
                    f"{task} Canaries={len(tg.PlacedCanaries or [])}/{tg.DesiredCanaries} Placed={tg.PlacedAllocs} Desired={tg.DesiredTotal} Healthy={tg.HealthyAllocs} Unhealthy={tg.UnhealthyAllocs} {deployment.StatusDescription}",
                )

    def stop(self):
        for w in self.workers.values():
            for th in w.taskhandlers.values():
                th.stop()

    def join(self):
        # Logs stream outputs empty {} which allows to handle timeouts.
        threads: List[Tuple[str, threading.Thread]] = [
            (f"{tk.task}[{i}]", logger)
            for w in self.workers.values()
            for tk, th in w.taskhandlers.items()
            for i, logger in enumerate(th.loggers or [])
        ]
        thcnt = sum(len(w.taskhandlers) for w in self.workers.values())
        log.debug(
            f"Joining {len(self.workers)} allocations with {thcnt} taskhandlers and {len(threads)} loggers"
        )
        timeend = time.time() + args.shutdown_timeout
        for desc, thread in threads:
            timeout = timeend - time.time()
            if timeout > 0:
                log.debug(f"joining worker {desc} timeout={timeout}")
                thread.join(timeout=timeout)
            else:
                log.debug("timeout passed for joining workers")
                break

    def exitcode(self) -> int:
        exitcodes: List[int] = [
            # If thread did not return, exit with -1.
            -1 if th.exitcode is None else th.exitcode
            for w in self.workers.values()
            for th in w.taskhandlers.values()
        ]
        if len(exitcodes) == 0:
            return ExitCode.no_allocations
        any_unfinished = any(v == -1 for v in exitcodes)
        if any_unfinished:
            return ExitCode.any_unfinished_tasks
        only_one_task = len(exitcodes) == 1
        if only_one_task:
            return exitcodes[0]
        all_failed = all(v != 0 for v in exitcodes)
        if all_failed:
            return ExitCode.all_failed_tasks
        any_failed = any(v != 0 for v in exitcodes)
        if any_failed:
            return ExitCode.any_failed_tasks
        return ExitCode.success


###############################################################################


def __nomad_job_run(opts: Tuple[str]) -> str:
    """Call nomad job run to start a Nomad job using nomad job run call with specified arguments"""
    cmd: List[str] = "nomad job run -detach -verbose".split() + list(opts)
    log.info(f"+ {' '.join(shlex.quote(x) for x in cmd)}")
    try:
        output = subprocess.check_output(cmd, text=True)
    except subprocess.CalledProcessError as e:
        # nomad will print its error, we can just exit
        exit(e.returncode)
    # Extract evaluation id from Nomad output.
    for line in output.splitlines():
        log.info(line)
    founduuids = re.findall(
        "[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}",
        output.lower(),
        re.MULTILINE,
    )
    assert len(founduuids) == 1, f"Could not find uuid in nomad job run output"
    evalid = founduuids[0]
    return evalid


def nomad_start_job(opts: Tuple[str]) -> nomadlib.Eval:
    """Start a nomad job using nomad job run parameters. Return evluation from running the job"""
    evalid = __nomad_job_run(opts)
    evaluation = nomadlib.Eval(mynomad.get(f"evaluation/{evalid}"))
    mynomad.namespace = evaluation.Namespace
    return evaluation


###############################################################################


class NomadJobWatcher(ABC):
    """Watches over a job. Schedules watches over allocations. Spawns loggers."""

    def __init__(
        self,
        jobid: Optional[str],
        eval: Optional[nomadlib.Eval],
        endstatusstr: str,
    ):
        assert (jobid and not eval) or (not jobid and eval)
        assert mynomad.namespace
        self.eval = eval
        """Evaluation or None"""
        self.endstatusstr = endstatusstr
        """Notify user up-until we are watching the job"""
        self.jobid: str = self.eval.JobID if self.eval else jobid if jobid else ""
        assert self.jobid
        """Watched job ID"""
        self.jobmodifyindex = self.eval.JobModifyIndex if self.eval else None
        """Watching job ID from this JobModifyIndex"""
        self.job: Optional[nomadlib.Job] = None
        """Last valid Job with ID"""
        self.db = NomadDbJob(
            topics=[
                f"Job:{self.jobid}",
                f"Evaluation:{self.jobid}",
                f"Allocation:{self.jobid}",
                f"Deployment:{self.jobid}",
            ],
            select_event_cb=self.db_select_event_job,
            init_cb=self.db_init_cb,
            force_polling=True if args.polling else None,
        )
        """Database listening to Nomad stream"""
        self.notifier = NotifierWorker(self.db)
        """Notification worker dispatching Nomad stream events"""
        self.done = threading.Event()
        """I am using threading.Event because you can't handle KeyboardInterrupt while Thread.join()."""
        self.thread = threading.Thread(
            target=self.__thread_run,
            name=f"{self.__class__.__name__}({self.jobid})",
            daemon=True,
        )
        self.dbseenjob: bool = False
        """The job was found at least once."""
        self.purged: bool = False
        """If set, we are waiting until the job is there and it means that JobNotFound is not an error - the job was removed."""
        self.purgedlock: threading.Lock = threading.Lock()
        """A special lock to synchronize callback with changing finish conditions"""
        self.donemsg = ""
        """Message printed when done watching. It is not printed right away, because we might decide to wait for purging later."""
        self.started: bool = False
        """The job finished because all allocations started, not because they failed."""
        #
        self.db.start()
        self.thread.start()

    def was_purged(self):
        with self.purgedlock:
            return self.purged

    def db_init_cb(self) -> List[Event]:
        """Db initialization callback"""
        try:
            job: dict = mynomad.get(f"job/{self.jobid}")
            jobversions: dict = mynomad.get(f"job/{self.jobid}/versions")["Versions"]
            evaluations: List[dict] = mynomad.get(f"job/{self.jobid}/evaluations")
            allocations: List[dict] = mynomad.get(f"job/{self.jobid}/allocations")
            deployments: List[dict] = mynomad.get(f"job/{self.jobid}/deployments")
        except nomadlib.JobNotFound:
            # This is fine to fail if it was purged, potentially.
            if self.was_purged() and self.job:
                # If the job was purged, generate one event so that listener can catch it.
                return [
                    Event(EventTopic.Job, EventType.JobRegistered, self.job.asdict())
                ]
            raise
        # Set the job if not set.
        if not self.job:
            self.job = nomadlib.Job(job)
            log.debug(
                f"Found job {self.job.description()} from {self.eval.ID if self.eval else None}"
            )
        if not allocations:
            log.debug(f"Job {self.job.description()} has no allocations")
        if self.jobmodifyindex is None:
            # set jobmodifyindex
            self.jobmodifyindex = self.job.JobModifyIndex
            # If there are active alocations with JobModifyIndex lower than current watched one,
            # also start watching them by lowering JobModfyIndex threashold.
            minactiveallocationjobmodifyindex = min(
                (
                    next((e for e in evaluations if e["ID"] == a["EvalID"]), {}).get(
                        "JobModifyIndex", self.jobmodifyindex
                    )
                    for a in allocations
                    if nomadlib.Alloc(a).is_pending_or_running()
                ),
                default=self.jobmodifyindex,
            )
            self.jobmodifyindex = min(
                self.jobmodifyindex, minactiveallocationjobmodifyindex
            )
            # Set the JobModifyIndex to the last still running Job version.
            runningjobs = [job for job in jobversions if not job["Stop"]]
            if runningjobs:
                lastrunningjob = sorted(
                    runningjobs, key=lambda job: job["ModifyIndex"]
                )[-1]
                self.jobmodifyindex = min(
                    self.jobmodifyindex, lastrunningjob["JobModifyIndex"]
                )
        # Finally output the events for database to pick up.
        return [
            *[
                Event(EventTopic.Deployment, EventType.DeploymentStatusUpdate, d)
                for d in deployments
            ],
            *[
                Event(EventTopic.Evaluation, EventType.EvaluationUpdated, e)
                for e in evaluations
            ],
            Event(EventTopic.Job, EventType.JobRegistered, job),
            *[
                Event(EventTopic.Allocation, EventType.AllocationUpdated, a)
                for a in allocations
            ],
            *[
                Event(EventTopic.Job, EventType.JobRegistered, job)
                for job in jobversions
            ],
        ]

    def db_select_event_jobid(self, e: Event):
        return self.db.apply_selects(
            e,
            lambda job: job.ID == self.jobid,
            lambda eval: eval.JobID == self.jobid,
            lambda alloc: alloc.JobID == self.jobid,
            lambda deploy: deploy.JobID == self.jobid,
        )

    def db_select_event_job(self, e: Event) -> bool:
        if args.all:
            return True
        job_filter: Callable[[nomadlib.Job], bool] = lambda _: True
        eval_filter: Callable[[nomadlib.Eval], bool] = lambda eval: (
            (self.eval is not None and self.eval.ID == eval.ID)
            # The JobModifyIndex has to be greater.
            or eval.get("JobModifyIndex", -1) >= self.jobmodifyindex
        )
        alloc_filter: Callable[[nomadlib.Alloc], bool] = lambda alloc: (
            # If allocation has JobVersion, then it has to match the version in the job.
            (self.job and alloc.get("JobVersion", -1) >= self.job.Version)
            or (
                # If the allocation has no JobVersion, find the maching evaluation.
                # The JobModifyIndex from the evaluation has to match.
                self.db.evaluations.get(alloc.EvalID, {}).get("JobModifyIndex", -1)
                >= self.jobmodifyindex
            )
        )
        deploy_filter: Callable[[nomadlib.Deploy], bool] = lambda deploy: (
            deploy.get("JobModifyIndex", -1) >= self.jobmodifyindex
        )
        return self.db_select_event_jobid(e) and self.db.apply_selects(
            e, job_filter, eval_filter, alloc_filter, deploy_filter
        )

    def finish_cb(self) -> bool:
        """Overloaded callback to call to determine if we should finish watching the job"""
        # Default is to return False and fallback to job_is_finished.
        return False

    def __thread_run(self):
        untilstr = (
            "forever"
            if args.all
            else f"for {args.shutdown_timeout} seconds"
            if args.no_follow
            else f"until it is {self.endstatusstr}"
        )
        log.info(f"Watching job {self.jobid}@{mynomad.namespace} {untilstr}")
        #
        no_follow_timeend = time.time() + args.shutdown_timeout
        for events in self.db.events():
            for event in events:
                if event.topic == EventTopic.Allocation:
                    alloc = nomadlib.Alloc(event.data)
                    self.notifier.notify_alloc(alloc)
                elif event.topic == EventTopic.Evaluation:
                    evaluation = nomadlib.Eval(event.data)
                    self.notifier.notify_eval(evaluation)
                    if self.eval and self.eval.ID == evaluation.ID:
                        self.eval = evaluation
                elif event.topic == EventTopic.Deployment:
                    deployment = nomadlib.Deploy(event.data)
                    self.notifier.notify_deploy(deployment)
                elif event.topic == EventTopic.Job:
                    if self.db.job:
                        self.dbseenjob = True
                    if (
                        self.job
                        and self.db.job
                        and self.job.ModifyIndex < self.db.job.ModifyIndex
                    ):
                        # self.job follows newest job definition.
                        self.job = self.db.job
            if (
                not self.done.is_set()
                and not args.follow
                # Evaluation is finished if was passed.
                and (self.eval is None or not self.eval.is_pending())
                and (
                    (self.finish_cb() or self.job_is_finished())
                    or (args.no_follow and time.time() > no_follow_timeend)
                )
            ):
                self.done.set()
        log.debug(f"Watching job {self.jobid}@{mynomad.namespace} exiting")

    @abstractmethod
    def _get_exitcode_cb(self) -> int:
        raise NotImplementedError()

    def get_exitcode(self) -> int:
        assert self.db.stopevent.is_set(), "stop not called"
        assert self.done.is_set(), "stop not called"
        return self._get_exitcode_cb()

    def wait(self):
        self.done.wait()

    def stop_job(self, purge: bool):
        self.db.initialized.wait()
        if purge:
            with self.purgedlock:
                self.purged = True
                self.done.clear()
        mynomad.stop_job(self.jobid, purge)

    def stop_threads(self):
        log.debug(f"stopping {self.__class__.__name__} done={self.done.is_set()}")
        self.notifier.stop()
        self.db.stop()
        self.notifier.join()
        self.db.join()
        mynomad.session.close()
        if self.donemsg:
            log.info(self.donemsg)

    def run_and_exit(self):
        try:
            self.wait()
        finally:
            self.stop_threads()
        exit(self.get_exitcode())

    def has_active_deployments(self):
        for deployments in self.db.deployments.values():
            if deployments.Status in [
                nomadlib.DeploymentStatus.initializing,
                nomadlib.DeploymentStatus.running,
                nomadlib.DeploymentStatus.pending,
                nomadlib.DeploymentStatus.blocked,
                nomadlib.DeploymentStatus.paused,
            ]:
                return True
        return False

    def has_active_evaluations(self):
        for evaluation in self.db.evaluations.values():
            if evaluation.is_pending():
                return True
        return False

    def has_no_active_allocations_and_evaluations_and_deployments(self):
        for allocation in self.db.allocations.values():
            if allocation.is_pending_or_running():
                return False
        return not self.has_active_deployments() and not self.has_active_evaluations()

    def _job_is_dead_message(self):
        assert self.job
        return f"Job {self.job.description()} is dead with no active allocations, evaluations nor deployments."

    def job_is_finished(self) -> bool:
        """Return True if the job is finished and nothing more will happen to it. Sets donemsg"""
        # Protect against the situation when the job JSON is not in a database.
        # init_cb first queries allocations, then the job itself.
        if (
            self.dbseenjob
            and self.job
            and self.has_no_active_allocations_and_evaluations_and_deployments()
        ):
            # Depending on purge argument, we wait for the job to stop existing
            # or for the job to be dead.
            if self.was_purged():
                if self.db.job is None:
                    self.donemsg = f"Job {self.job.description()} removed with no active allocations, evaluations nor deployments. Exiting."
                    return True
            else:
                if self.job.is_dead():
                    self.donemsg = f"{self._job_is_dead_message()} Exiting."
                    return True
        return False

    def is_current_allocation(self, alloc: nomadlib.Alloc) -> bool:
        """Return True if the allocation is for the current job version"""
        return self.job is not None and (
            alloc.get("JobVersion", -1) >= self.job.Version
            or self.db.evaluations.get(alloc.EvalID, {}).get("JobModifyIndex", -1)
            >= self.job.JobModifyIndex
        )

    @staticmethod
    def nomad_job_group_main_tasks(group: nomadlib.JobTaskGroup):
        """Get a set of names of Nomad job group main tasks"""
        # Main tasks are tasks that:
        # - do not have lifecycle
        # - have lifecycle prestart with sidecar = true
        # - have lifecycle poststart
        # All these tasks have to be started.
        maintasks = set(
            t.Name
            for t in group.Tasks
            if "Lifecycle" not in t
            or t.Lifecycle is None
            or (
                t.Lifecycle.Hook == nomadlib.LifecycleHook.prestart
                and t.Lifecycle.get_sidecar() == True
            )
            or t.Lifecycle.Hook == nomadlib.LifecycleHook.poststart
        )
        assert len(maintasks) > 0, f"Internal error when getting main tasks of {group}"
        return maintasks

    def job_has_finished_starting(self) -> bool:
        """
        Return True if the job is not starting anymore.
        self.started will be set to True, if the job was successfully started.
        """
        # log.debug(f"has_active_deployments={self.has_active_deployments()} has_active_evaluations={self.has_active_evaluations()} allocations={len(self.db.allocations)}")
        if (
            not self.job
            or self.has_active_deployments()
            or self.has_active_evaluations()
        ):
            return False
        allocations = self.db.allocations.values()
        if not allocations:
            return False
        groupmsgs: List[str] = []
        for group in self.job.TaskGroups:
            # Similar logic in db_filter_event_job()
            groupallocs: List[nomadlib.Alloc] = [
                alloc
                for alloc in allocations
                if alloc.TaskGroup == group.Name and self.is_current_allocation(alloc)
            ]
            if len(groupallocs) != group.Count:
                # This group has no current allocations,
                # and no active evaluation and deployments (checked above).
                log.info(
                    f"Job {self.job.description()} has failed to start group {group.Name!r}"
                )
                return True
            maintasks: Set[str] = self.nomad_job_group_main_tasks(group)
            for alloc in groupallocs:
                # List of started tasks.
                startedtasks: Set[str] = set(
                    name
                    for name, taskstate in alloc.get_taskstates().items()
                    if taskstate.was_started()
                )
                notrunningmaintasks = maintasks.difference(startedtasks)
                if notrunningmaintasks:
                    # There are main tasks that are not running.
                    if alloc.is_pending_or_running():
                        return False
                    else:
                        # The allocation has finished - the tasks will never start.
                        log.info(
                            f"Job {self.job.description()} has failed to start group {group.Name!r} tasks {' '.join(notrunningmaintasks)}"
                        )
                        return True
            groupallocsidsstr = andjoin(alloc.ID[:6] for alloc in groupallocs)
            groupmsgs.append(
                f"allocations {groupallocsidsstr} running group {group.Name!r} with {len(maintasks)} main tasks"
            )
        msg = f"Job {self.job.description()} started " + andjoin(groupmsgs) + "."
        log.info(msg)
        self.started = True
        return True


class NomadJobWatcherUntilFinished(NomadJobWatcher):
    """Watcher a job until the job is dead or purged"""

    def __init__(
        self,
        jobid: Optional[str],
        eval: Optional[nomadlib.Eval] = None,
    ):
        super().__init__(jobid, eval, "finished")

    def _get_exitcode_cb(self) -> int:
        exitcode: int = (
            (ExitCode.success if self.done.is_set() else ExitCode.interrupted)
            if args.no_preserve_status
            else self.notifier.exitcode()
        )
        log.debug(f"exitcode={exitcode}")
        return exitcode

    def job_finished_successfully(self):
        self.db.initialized.wait()
        assert self.job
        if self.job.Status != "dead":
            return False
        s: nomadlib.JobSummarySummary = nomadlib.JobSummary(
            mynomad.get(f"job/{self.job.ID}/summary")
        ).get_sum_summary()
        log.debug(f"{self.job_finished_successfully.__name__}: {s}")
        return (
            s.Queued == 0
            and s.Complete != 0
            and s.Failed == 0
            and s.Running == 0
            and s.Starting == 0
            and s.Lost == 0
        )

    def job_running_successfully(self):
        self.db.initialized.wait()
        assert self.job
        if self.job.Status == "dead":
            return self.job_finished_successfully()
        s: nomadlib.JobSummarySummary = nomadlib.JobSummary(
            mynomad.get(f"job/{self.jobid}/summary")
        ).get_sum_summary()
        log.debug(f"{self.job_running_successfully.__name__} {s}")
        return (
            s.Queued == 0
            and s.Failed == 0
            and s.Running != 0
            and s.Starting == 0
            and s.Lost == 0
        )


class NomadJobWatcherUntilStarted(NomadJobWatcher):
    """Watches a job until the job is started"""

    def __init__(
        self,
        jobid: Optional[str],
        eval: Optional[nomadlib.Eval] = None,
    ):
        super().__init__(jobid, eval, "started")

    def finish_cb(self) -> bool:
        return self.job_has_finished_starting()

    def _get_exitcode_cb(self) -> int:
        exitcode = (
            ExitCode.interrupted
            if not self.done.is_set()
            else ExitCode.failed
            if not self.started
            else ExitCode.success
        )
        log.debug(
            f"done={self.done.is_set()} started={self.started} exitcode={exitcode}"
        )
        return exitcode


class NomadAllocationWatcher:
    """Watch an allocation until it is finished"""

    def __init__(self, alloc: nomadlib.Alloc):
        self.alloc = alloc
        self.db = NomadDbJob(
            topics=[f"Allocation:{alloc.ID}"],
            select_event_cb=lambda e: e.topic == EventTopic.Allocation
            and e.data["ID"] == alloc.ID,
            init_cb=lambda: [
                Event(
                    EventTopic.Allocation,
                    EventType.AllocationUpdated,
                    mynomad.get(f"allocation/{alloc.ID}"),
                )
            ],
            force_polling=True if args.polling else None,
        )
        self.allocworkers = NotifierWorker(self.db)
        self.finished = False
        #
        log.info(f"Watching allocation {alloc.ID}")
        self.db.start()

    def run_and_exit(self):
        try:
            for events in self.db.events():
                for event in events:
                    if event.topic == EventTopic.Allocation:
                        alloc = nomadlib.Alloc(event.data)
                        assert (
                            self.alloc.ID == alloc.ID
                        ), f"Internal error in Db filter: {alloc} {self.alloc}"
                        self.allocworkers.notify_alloc(alloc)
                        if alloc.is_finished():
                            log.info(
                                f"Allocation {alloc.ID} has status {alloc.ClientStatus}. Exiting."
                            )
                            self.finished = True
                            break
                if args.no_follow or self.finished:
                    break
        finally:
            self.db.stop()
            self.allocworkers.stop()
            self.db.join()
            self.allocworkers.join()
        exit(self.allocworkers.exitcode())


###############################################################################


class JobPath:
    """Unused currently"""

    jobname: str
    group: Optional[Pattern[str]]
    task: Optional[Pattern[str]]

    def __init__(self, param: str):
        a = param.split("@")
        self.jobname = a[0]
        if len(a) == 2:
            self.task = re.compile(a[1])
        elif len(a) == 3:
            self.group = re.compile(a[1])
            self.task = re.compile(a[2])
        assert (
            1 <= len(a) <= 3
        ), f"Invalid job/job@task/job@group@task specification: {param}"

    @staticmethod
    def complete(ctx: click.Context, _: str, incomplete: str) -> List[str]:
        _complete_set_namespace(ctx)
        try:
            jobs = mynomad.get("jobs")
        except requests.HTTPError:
            return []
        jobsids = [x["ID"] for x in jobs]
        arg = incomplete.split("@")
        complete = []
        if len(arg) == 1:
            complete = [x for x in jobsids]
            complete += [f"{x}@" for x in jobsids]
        elif len(arg) == 2 or len(arg) == 3:
            jobid = arg[0]
            jobsids = [x for x in jobsids if x == arg[0]]
            if len(jobsids) != 1:
                return []
            mynomad.namespace = next(x for x in jobs if x["ID"] == arg[0])["Namespace"]
            try:
                job = nomadlib.Job(mynomad.get(f"job/{jobid}"))
            except requests.HTTPError:
                return []
            if len(arg) == 2:
                tasks = [t.Name for tg in job.TaskGroups for t in tg.Tasks]
                groups = [tg.Name for tg in job.TaskGroups]
                complete = itertools.chain(tasks, groups)
            elif len(arg) == 3:
                complete = [f"{tg.Name}@{t}" for tg in job.TaskGroups for t in tg.Tasks]
            complete = [f"{arg[0]}@{x}" for x in complete]
        else:
            return []
        return [x for x in complete if x.startswith(incomplete)]


@click.group(
    help=f"""
Depending on the command, run or stop a Nomad job. Watch over the job and
print all job allocation events and tasks stdouts and tasks stderrs
logs. Depending on command, wait for a specific event to happen to finish
watching. This program is intended to help debugging issues with running
jobs in Nomad and for synchronizing with execution of batch jobs in Nomad.

Logs are printed in the format: 'mark>id>#version>group>task> message'.
The mark in the log lines is equal to: 'deploy' for messages printed as
a result of deployment, 'eval' for messages printed from evaluations,
'A' from allocation, 'E' for stderr logs of a task and 'O' from stdout
logs of a task.

\b
Examples:
    nomad-watch run ./some-job.nomad.hcl
    nomad-watch job some-job
    nomad-watch alloc af94b2
    nomad-watch -N services --task redis -1f job redis
""",
    epilog="""
    Written by Kamil Cukrowski 2023. Licensed under GNU GPL version 3 or later.
    """,
)
@namespace_option()
@click.option(
    "-a",
    "--all",
    is_flag=True,
    help="Print logs from all allocations, including previous versions of the job.",
)
@click.option(
    "-o",
    "--out",
    type=click.Choice(
        "all alloc A stdout out O 1 stderr err E 2 evaluation eval e deployment deploy d none".split()
    ),
    default=["all"],
    multiple=True,
    show_default=True,
    help="Choose which stream of messages to print - evaluation, allocation, stdout, stderr. This option is cumulative.",
)
@click.option("-v", "--verbose", count=True, help="Be more verbose.")
@click.option("-q", "--quiet", count=True, help="Be less verbose.")
@click.option(
    "-A",
    "--attach",
    is_flag=True,
    help="Stop the job on interrupt and after it has finished. Relevant in run mode only.",
)
@click.option(
    "--purge-successful",
    is_flag=True,
    help="""
        When stopping the job, purge it when all job summary metrics are zero except nonzero complete metric.
        Relevant in run and stop modes. Implies --attach.
        """,
)
@click.option(
    "--purge",
    is_flag=True,
    help="When stopping the job, purge it. Relevant in run and stop modes. Implies --attach.",
)
@click.option(
    "-n",
    "--lines",
    default=10,
    show_default=True,
    type=int,
    help="""
        Sets the tail location in best-efforted number of lines relative to the end of logs.
        Negative value prints all available log lines.
        """,
)
@click.option(
    "--lines-timeout",
    default=0.5,
    show_default=True,
    type=float,
    help="When using --lines the number of lines is best-efforted by ignoring lines for this specific time",
)
@click.option(
    "--shutdown-timeout",
    default=2,
    show_default=True,
    type=float,
    help="The time to wait to make sure task loggers received all logs when exiting.",
)
@click.option(
    "-f",
    "--follow",
    is_flag=True,
    help="Never exit",
)
@click.option(
    "--no-follow",
    is_flag=True,
    help="Just run once, get the logs in a best-effort style and exit.",
)
@click.option(
    "-t",
    "--task",
    type=re.compile,
    help="Only watch tasks names matching this regex.",
)
@click.option(
    "--polling",
    is_flag=True,
    help="Instead of listening to Nomad event stream, periodically poll for events.",
)
@click.option(
    "-x",
    "--no-preserve-status",
    is_flag=True,
    help="Do not preserve tasks exit statuses.",
)
@click_log_options()
@common_options()
def cli(**kwargs):
    exit_on_thread_exception.install()
    global args
    args = argparse.Namespace(**kwargs)
    assert (
        (not args.follow and not args.no_follow)
        or (args.follow and not args.no_follow)
        or (not args.follow and args.no_follow)
    ), f"--follow and --no-follow conflict"
    #
    global START_NS
    START_NS = time.time_ns()
    init_logging()


cli_jobid = click.argument(
    "jobid",
    shell_complete=complete_job(),
)


def cli_jobfile(name: str, help: str):
    # Disabled, because maintenance.
    nomad_job_run_flags = [
        click.option("-check-index", type=int),
        click.option("-detach", is_flag=True),
        click.option("-eval-priority", type=int),
        click.option("-json", is_flag=True),
        click.option("-hcl1", is_flag=True),
        click.option("-hcl2-strict", is_flag=True),
        click.option("-policy-override", is_flag=True),
        click.option("-preserve-counts", is_flag=True),
        click.option("-consul-token"),
        click.option("-vault-token"),
        click.option("-vault-namespace"),
        click.option("-var", multiple=True),
        click.option("-var-file", type=click.File()),
    ]
    cli_jobfile_help = """
    JOBFILE can be file with a HCL or JSON nomad job or
    it can be a string containing a HCL or JSON nomad job.
    """
    return composed(
        cli.command(
            name,
            help + cli_jobfile_help,
            context_settings=dict(ignore_unknown_options=True),
        ),
        *nomad_job_run_flags,
        click.argument(
            "jobfile",
            shell_complete=click.File().shell_complete,
        ),
    )


def cli_command_run_nomad_job_run(name: str, help: str):
    return composed(
        cli.command(
            name,
            help=help.rstrip()
            + """\
            All following command arguments are passed to nomad job run command.
            Note that nomad job run has arguments with a single dash.
            """,
            context_settings=dict(ignore_unknown_options=True),
        ),
        click.argument(
            "cmd",
            nargs=-1,
            shell_complete=click.File().shell_complete,
        ),
    )


###############################################################################


@cli.command(
    "alloc",
    help="Watch over specific allocation. Like job mode, but only one allocation is filtered.",
)
@click.argument(
    "allocid",
    shell_complete=completor(lambda: (x["ID"] for x in mynomad.get("allocations"))),
)
@common_options()
def mode_alloc(allocid):
    allocs = mynomad.get(f"allocations", params={"prefix": allocid})
    assert len(allocs) > 0, f"Allocation with id {allocid} not found"
    assert len(allocs) < 2, f"Multiple allocations found starting with id {allocid}"
    alloc = nomadlib.Alloc(allocs[0])
    mynomad.namespace = alloc.Namespace
    NomadAllocationWatcher(alloc).run_and_exit()


@cli.command(
    "eval", help="Watch like job mode the job that results from a specific evaluation."
)
@click.argument(
    "evalid",
    shell_complete=completor(lambda: (x["ID"] for x in mynomad.get("evaluations"))),
)
@common_options()
def mode_eval(evalid):
    evaluation = mynomad.get(f"evaluations", params={"prefix": evalid})
    assert len(evaluation) > 0, f"Evaluation with id {evalid} not found"
    assert len(evaluation) < 2, f"Multiple evaluations found starting with id {evalid}"
    NomadJobWatcherUntilFinished(None, evaluation).run_and_exit()


@cli_command_run_nomad_job_run(
    "run",
    help=f"Run a Nomad job and then act like started mode.",
)
def mode_run(cmd: Tuple[str]):
    evaluation = nomad_start_job(cmd)
    do = NomadJobWatcherUntilFinished(None, evaluation)
    try:
        do.wait()
    finally:
        # On normal execution, the job is stopped.
        # On KeyboardException, job is still running.
        if args.attach or args.purge or args.purge_successful:
            purge: bool = args.purge or (
                args.purge_successful and do.job_finished_successfully()
            )
            do.stop_job(purge)
            do.wait()
        do.stop_threads()
    exit(do.get_exitcode())


@cli.command(
    "job",
    help="Alias to stopped command.",
)
@cli_jobid
@common_options()
def mode_job(jobid: str):
    jobid = nomad_find_job(jobid)
    NomadJobWatcherUntilFinished(jobid).run_and_exit()


@cli_command_run_nomad_job_run(
    "start", help=f"Start a Nomad Job and then act like started command."
)
def mode_start(cmd: Tuple[str]):
    evaluation = nomad_start_job(cmd)
    NomadJobWatcherUntilStarted(None, evaluation).run_and_exit()


@cli.command(
    "started",
    help=f"""
Watch a Nomad job until the job is started.
Job is started when it has no active deployments
and no active evaluations
and the number of allocations is equal to the number of groups multiplied by group count
and all main tasks in each allocation are running.
An active deployment is a deployment that has status equal to
initializing, running, pending, blocked or paused.
Main tasks are all tasks without lifetime property or sidecar prestart tasks or poststart tasks.

\b
Exit with the following status:
  {ExitCode.success    }  when all tasks of the job have started running,
  {ExitCode.exception  }  when python exception was thrown,
  {ExitCode.interrupted}  when process was interrupted,
  {ExitCode.failed     }  when job was stopped or job deployment was reverted.
""",
)
@cli_jobid
@common_options()
def mode_started(jobid: str):
    jobid = nomad_find_job(jobid)
    NomadJobWatcherUntilStarted(jobid).run_and_exit()


def mode_stop_in(jobid: str):
    jobid = nomad_find_job(jobid)
    do = NomadJobWatcherUntilFinished(jobid)
    purge: bool = args.purge or (
        args.purge_successful and do.job_running_successfully()
    )
    do.stop_job(purge)
    do.run_and_exit()


@cli.command(
    "stop",
    help="Stop a Nomad job and then act like stopped command.",
)
@cli_jobid
@common_options()
def mode_stop(jobid: str):
    mode_stop_in(jobid)


@cli.command(
    "purge",
    help="Alias to --purge stop.",
)
@cli_jobid
@common_options()
def mode_purge(jobid: str):
    args.purge = True
    mode_stop_in(jobid)


@cli.command(
    "stopped",
    help=f"""
Watch a Nomad job until the job is stopped.
Job is stopped when the job is dead or, if the job was purged, does not exists anymore,
and the job has no running or pending allocations,
no active deployments and no active evaluations.

\b
If the option --no-preserve-status is given, then exit with the following status:
  {ExitCode.success           }    when the job was stopped
Otherwise, exit with the following status:
  {'?'                        }    when the job has one task, with that task exit status,
  {ExitCode.success           }    when all tasks of the job exited with 0 exit status,
  {ExitCode.any_failed_tasks    }  when any of the job tasks have failed,
  {ExitCode.all_failed_tasks    }  when all job tasks have failed,
  {ExitCode.any_unfinished_tasks}  when any tasks are still running,
  {ExitCode.no_allocations      }  when job has no started tasks.
In any case, exit with the following exit status:
  {ExitCode.exception         }    when python exception was thrown,
  {ExitCode.interrupted       }    when the process was interrupted.
""",
)
@cli_jobid
@common_options()
def mode_stopped(jobid: str):
    jobid = nomad_find_job(jobid)
    NomadJobWatcherUntilFinished(jobid).run_and_exit()


###############################################################################

if __name__ == "__main__":
    try:
        cli.main()
    finally:
        mynomad.session.close()



import json
import logging
import queue
import threading
from typing import Callable, Dict, Iterable, List, Optional, TypeVar, Union

import requests

from . import nomadlib
from .common import json_loads, mynomad
from .nomadlib import Event, EventTopic, EventType

log = logging.getLogger(__name__)
T = TypeVar("T")


class NomadDbJob:
    """Represents relevant state cache from Nomad database of a single job"""

    def __init__(
        self,
        topics: List[str],
        select_event_cb: Callable[[Event], bool],
        init_cb: Callable[[], List[Event]],
        force_polling: Optional[bool] = None,
    ):
        """
        :param topic The topics to listen to, see Nomad event stream API documentation.
        :param select_event_cb: Filter only relevant events from Nomad event stream.
        :param init_cb Return a list of events to populate the database with and to poll. Has to be threadsafe!
        """
        self.topics: List[str] = topics
        self.init_cb: Callable[[], List[Event]] = init_cb
        self.select_event_cb: Callable[[Event], bool] = select_event_cb
        self.force_polling = force_polling
        self.queue: queue.Queue[Optional[List[Event]]] = queue.Queue()
        """Queue where database thread puts the received events"""
        self.job: Optional[nomadlib.Job] = None
        """Watched job"""
        self.jobversions: Dict[int, nomadlib.Job] = {}
        """Job version to job definition"""
        self.evaluations: Dict[str, nomadlib.Eval] = {}
        """Database of evaluations"""
        self.allocations: Dict[str, nomadlib.Alloc] = {}
        """Database of allocations"""
        self.deployments: Dict[str, nomadlib.Deploy] = {}
        """Database of deployments"""
        self.initialized = threading.Event()
        """Was init_cb called to initiliaze the database"""
        self.stopevent = threading.Event()
        """If set, the database thread should exit"""
        self.thread = threading.Thread(
            target=self.__thread_entry, name="db", daemon=True
        )
        assert self.topics
        assert not any(not x for x in topics)

    ###############################################################################

    def __thread_run_stream(self):
        log.debug(f"Starting listen Nomad stream with {' '.join(self.topics)}")
        with mynomad.stream(
            "event/stream",
            params={"topic": self.topics},
        ) as stream:
            for line in stream.iter_lines():
                if line:
                    data = json_loads(line)
                    events: List[Event] = [
                        Event(
                            EventTopic[event["Topic"]],
                            EventType[event["Type"]],
                            event["Payload"][event["Topic"]],
                            stream=True,
                        )
                        for event in data.get("Events", [])
                    ]
                    # log.debug(f"RECV EVENTS: {events}")
                    self.queue.put(events)
                if self.stopevent.is_set():
                    break

    def __thread_poll(self):
        """If listening to stream fails, we fallback to calling init_cb in a loop"""
        # Delay polling start up until init_cb has been called.
        self.initialized.wait()
        while not self.stopevent.wait(1):
            self.queue.put(self.init_cb())

    def __thread_entry(self):
        """Database thread entry"""
        try:
            try:
                if self.force_polling is True:
                    self.__thread_poll()
                else:
                    try:
                        self.__thread_run_stream()
                    except nomadlib.PermissionDenied as e:
                        if self.force_polling is False:
                            raise
                        else:
                            log.warning(
                                f"Falling to polling method because stream API returned permission denied: {e}"
                            )
                    self.__thread_poll()
            finally:
                log.debug("Nomad database thread exiting")
                self.queue.put(None)
        except requests.HTTPError as e:
            log.exception("http request failed")
            exit(1)

    ###############################################################################

    def start(self):
        """Start the database thread"""
        assert (
            mynomad.namespace
        ), "Nomad namespace has to be set before starting to listen"
        self.thread.start()

    def select_is_in_db(self, e: Event):
        """Select events that are already in the database"""
        if e.topic == EventTopic.Evaluation:
            return e.data["ID"] in self.evaluations
        elif e.topic == EventTopic.Allocation:
            return e.data["ID"] in self.allocations
        elif e.topic == EventTopic.Deployment:
            return e.data["ID"] in self.deployments
        elif e.topic == EventTopic.Job:
            return e.data["Version"] in self.jobversions
        return False

    def _add_event_to_db(self, e: Event):
        """Update database state to reflect received event"""
        if e.topic == EventTopic.Job:
            job = nomadlib.Job(e.data)
            self.jobversions[job.get("Version")] = job
            self.job = None if e.type == EventType.JobDeregistered else job
        elif e.topic == EventTopic.Evaluation:
            if e.type == EventType.JobDeregistered:
                self.job = None
            self.evaluations[e.data["ID"]] = nomadlib.Eval(e.data)
        elif e.topic == EventTopic.Allocation:
            # Events from event stream are missing JobVersion. Try to preserve it here by preserving keys.
            self.allocations[e.data["ID"]] = nomadlib.Alloc(
                {**self.allocations.get(e.data["ID"], {}), **e.data}
            )
        elif e.topic == EventTopic.Deployment:
            self.deployments[e.data["ID"]] = nomadlib.Deploy(e.data)

    def handle_event(self, e: Event) -> bool:
        if self._select_new_event(e):
            if self.select_is_in_db(e) or self.select_event_cb(e):
                # log.debug(f"EVENT: {e}")
                # log.debug(f"EVENT: {e} {e.data}")
                self._add_event_to_db(e)
                return True
            else:
                # log.debug(f"USER FILTERED: {e}")
                pass
        else:
            # log.debug(f"OLD EVENT: {e}")
            pass
        return False

    def handle_events(self, events: List[Event]) -> List[Event]:
        """From a list of events, filter out ignored and add the rest to database"""
        return [e for e in events if self.handle_event(e)]

    @staticmethod
    def apply_selects(
        e: Event,
        job_select: Callable[[nomadlib.Job], bool],
        eval_select: Callable[[nomadlib.Eval], bool],
        alloc_select: Callable[[nomadlib.Alloc], bool],
        deploy_select: Callable[[nomadlib.Deploy], bool],
    ) -> bool:
        """Apply specific selectors depending on event type"""
        return e.apply(job_select, eval_select, alloc_select, deploy_select)

    def _select_new_event(self, e: Event):
        """Select events which are newer than those in the database"""
        job_select: Callable[[nomadlib.Job], bool] = lambda _: True
        eval_select: Callable[[nomadlib.Eval], bool] = (
            lambda eval: eval.ID not in self.evaluations
            or eval.ModifyIndex > self.evaluations[eval.ID].ModifyIndex
        )
        alloc_select: Callable[[nomadlib.Alloc], bool] = (
            lambda alloc: alloc.ID not in self.allocations
            or alloc.ModifyIndex > self.allocations[alloc.ID].ModifyIndex
        )
        deploy_select: Callable[[nomadlib.Deploy], bool] = (
            lambda deploy: deploy.ID not in self.deployments
            or deploy.ModifyIndex > self.deployments[deploy.ID].ModifyIndex
        )
        return e.data["Namespace"] == mynomad.namespace and self.apply_selects(
            e, job_select, eval_select, alloc_select, deploy_select
        )

    def stop(self):
        log.debug("Stopping listen Nomad stream")
        self.initialized.set()
        self.stopevent.set()

    def join(self):
        # Not joining - neither requests nor stream API allow for timeouts.
        # self.thread.join()
        pass

    def events(self) -> Iterable[List[Event]]:
        """Nomad stream returns Events array. Iterate over batches of events returned from Nomad stream"""
        assert self.thread.is_alive(), "Thread not alive"
        if not self.initialized.is_set():
            events = self.init_cb()
            events = self.handle_events(events)
            yield events
            self.initialized.set()
        log.debug("Starting getting events from thread")
        while not self.queue.empty() or (
            self.thread.is_alive() and not self.stopevent.is_set()
        ):
            events = self.queue.get()
            if events is None:
                break
            yield self.handle_events(events)
        log.debug("db exiting")

    def get_allocation_job_version(
        self, alloc: nomadlib.Alloc, default: T = None
    ) -> Union[T, int]:
        """Given an allocation return the job version associated with that allocation"""
        if "JobVersion" in alloc and alloc.JobVersion is not None:
            return alloc.JobVersion
        evaluation = self.evaluations.get(alloc.EvalID)
        if not evaluation:
            return default
        jobversion = self.find_jobversion_from_modifyindex(evaluation.JobModifyIndex)
        if jobversion is not None:
            alloc.JobVersion = jobversion
            return jobversion
        return default

    def find_job_from_modifyindex(self, jobmodifyindex: int) -> Optional[nomadlib.Job]:
        for _, job in sorted(self.jobversions.items(), reverse=True):
            if job.JobModifyIndex <= jobmodifyindex:
                return job
        return None

    def find_jobversion_from_modifyindex(self, jobmodifyindex: int) -> Optional[int]:
        ret = self.find_job_from_modifyindex(jobmodifyindex)
        return ret.Version if ret else None



# https://stackoverflow.com/questions/49663124/cause-python-to-exit-if-any-thread-has-an-exception

import os
import threading
import traceback

original_init = threading.Thread.__init__
threadexception = None


def patched_init(self, *args, **kwargs):
    original_init(self, *args, **kwargs)
    original_run = self.run

    def patched_run(*args, **kw):
        try:
            original_run(*args, **kw)
        except:
            traceback.print_exc()
            os._exit(1)

    self.run = patched_run


def install():
    threading.Thread.__init__ = patched_init



import json
import logging
import os
from typing import Any, Callable, Dict, Iterable

import click
import pkg_resources

from . import nomadlib

log = logging.getLogger(__name__)
mynomad = nomadlib.NomadConn()


def nomad_find_job(id: str) -> str:
    """Find job named jobprefix if namespace is *."""
    if os.environ["NOMAD_NAMESPACE"] != "*":
        mynomad.namespace = os.environ["NOMAD_NAMESPACE"]
        return id
    jobs = mynomad.get("jobs", params={"prefix": id})
    jobsstr = " ".join(f"{job['ID']}@{job['Namespace']}" for job in jobs)
    try:
        found = next(job for job in jobs if job["ID"] == id)
        mynomad.namespace = found["Namespace"]
        os.environ["NOMAD_NAMESPACE"] = found["Namespace"]
        return found
    except StopIteration:
        log.exception(f"There are no jobs named {id}, closest are {jobsstr}")
        raise


def _complete_set_namespace(ctx: click.Context):
    namespace = ctx.params.get("namespace")
    if namespace:
        os.environ["NOMAD_NAMESPACE"] = namespace


def completor(cb: Callable[[], Iterable[str]]):
    def completor_cb(ctx: click.Context, param: str, incomplete: str):
        _complete_set_namespace(ctx)
        try:
            return [x for x in cb() if x.startswith(incomplete)]
        except Exception:
            pass

    return completor_cb


def namespace_option():
    return click.option(
        "-N",
        "--namespace",
        help="Finds Nomad namespace matching given prefix and sets NOMAD_NAMESPACE environment variable.",
        envvar="NOMAD_NAMESPACE",
        show_default=True,
        default="default",
        shell_complete=completor(
            lambda: (x["Name"] for x in mynomad.get("namespaces"))
        ),
        callback=lambda ctx, param, value: os.environ.__setitem__(
            "NOMAD_NAMESPACE", value
        ),
    )


def complete_job():
    return completor(lambda: (x["ID"] for x in mynomad.get("jobs")))


def composed(*decs):
    """Merge decorators into one decorator"""

    def deco(f):
        for dec in reversed(decs):
            f = dec(f)
        return f

    return deco


def get_version():
    return pkg_resources.get_distribution(__package__).version


def __print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    # Copied from version_option()
    prog_name = ctx.find_root().info_name
    click.echo(f"{prog_name}, version {get_version()}")
    ctx.exit()


def common_options():
    return composed(
        click.help_option("-h", "--help"),
        click.option(
            "--version",
            is_flag=True,
            callback=__print_version,
            expose_value=False,
            is_eager=True,
            help="Print program version then exit.",
        ),
    )


def json_loads(txt: str) -> Any:
    try:
        return json.loads(txt)
    except json.JSONDecodeError:
        log.exception(f"Could not json.loads: {txt!r}")
        raise


###############################################################################


def __alias_option_callback(
    aliased: Dict[str, Any],
    ctx: click.Context,
    param: click.Parameter,
    value: Any,
):
    """Callback called from alias_option option."""
    if value:
        for paramname, val in aliased.items():
            param = next(p for p in ctx.command.params if p.name == paramname)
            param.default = val


def alias_option(
    *param_decls: str,
    aliased: Dict[str, Any],
    **attrs: Any,
):
    """Add this to click options to have an alias for other options"""
    aliasedhelp = " ".join(
        "--"
        + k.replace("_", "-")
        + ("" if v is True else f"={v}" if isinstance(v, int) else f"={v!r}")
        for k, v in aliased.items()
    )
    return click.option(
        *param_decls,
        is_flag=True,
        help=f"Alias to {aliasedhelp}",
        callback=lambda *args: __alias_option_callback(aliased, *args),
        **attrs,
    )



#!/usr/bin/env python3

import argparse
import distutils.spawn
import json
import logging
import os
import subprocess
import tempfile
from itertools import chain
from pathlib import Path
from shlex import quote, split
from typing import Dict

import click

from .common import common_options, complete_job, mynomad, namespace_option
from .nomadlib import VariableNotFound
from .nomadlib.connection import VariableConflict

log = logging.getLogger(__file__)

###############################################################################


def dryrunstr():
    return "DRYRUN: " if args.dryrun else ""


def get_namespace_job_from_nomad_service_file(file: Path):
    try:
        with file.open() as f:
            jobjson = json.load(f)
    except json.JSONDecodeError:
        jobjson = json.loads(
            subprocess.check_output(
                "nomad job run -output".split() + [str(file)], text=True
            )
        )
    return jobjson["Job"]["Namespace"], jobjson["Job"]["ID"]


def create_tree(dir: Path, data: Dict[str, str]):
    """Create files with pahts as keys from data and content as values from data in directory dir"""
    dir.mkdir(parents=True, exist_ok=True)
    for k, v in data.items():
        p = dir / k
        p.parent.mkdir(parents=True, exist_ok=True)
        with p.open("w+") as f:
            f.write(v)


###############################################################################


class NomadVarDirMain:
    def __init__(self, ctx):
        global args
        args = argparse.Namespace(**{**vars(args), **ctx.params})
        logging.basicConfig(
            format="%(module)s: %(message)s",
            level=logging.DEBUG if args.verbose else logging.INFO,
        )
        if args.path:
            assert not args.service, "--service can't be passed with --path"
            assert not args.job, "--job can't be passed with --path"
        elif args.service:
            assert not args.job, "--service can't be passed with --job"
            args.namespace, args.job = get_namespace_job_from_nomad_service_file(
                args.service
            )
            os.environ["NOMAD_NAMESPACE"] = args.namespace
        elif args.job:
            args.path = f"nomad/job/{args.job}"
        else:
            assert 0, "One of --path --service or --job has to be given."
        self.get_old_items()

    def get_old_items(self):
        try:
            olditems = mynomad.variables.read(args.path).Items
        except VariableNotFound:
            olditems = {}
        self.olditems: Dict[str, str] = olditems
        return olditems

    def gen_new_items(self):
        # Get all files and directories recursively passed by args.paths.
        files = []
        for path in args.paths:
            if path.is_dir():
                files += (x for x in path.rglob("*") if x.is_file())
            else:
                files += [path]
        # If args.relative, the paths are relative to the directory.
        if args.relative:
            os.chdir(args.relative)
            files = [x.relative_to(args.relative) for x in files]
        if not args.disable_size_check:
            limit_mb = 10
            for file in files:
                filesize_mb = int(file.stat().st_size / 1024 / 1024)
                assert (
                    filesize_mb < limit_mb
                ), f"{file} size is {filesize_mb} greater than {limit_mb} mb, exiting"
        self.newitems: Dict[str, str] = {str(file): file.read_text() for file in files}
        if "D" in args and args.D:
            assert all("=" in x for x in args.D), "-D options have to be var=value"
            self.newitems.update({k: v for x in args.D for k, v in x.split("=", 2)})
        if "clear" in args and not args.clear:
            self.newitems = {**self.olditems, **self.newitems}
        return self.newitems

    def show_diff_11(self):
        if not distutils.spawn.find_executable("diff"):
            log.warning("diff executable not found")
            return
        with tempfile.NamedTemporaryFile("w", prefix="nomad-var-dir_") as tmpf:
            for k in sorted(
                list(set(chain(self.newitems.keys(), self.olditems.keys())))
            ):
                old = self.olditems.get(k, "")
                tmpf.seek(0)
                tmpf.write(old)
                tmpf.flush()
                #
                newfile = (
                    (k if Path(k).exists() else "-")
                    if k in self.newitems
                    else subprocess.DEVNULL
                )
                stdin = self.newitems[k] if newfile == "-" else None
                subprocess.run(
                    split(f"diff --color -u {quote(tmpf.name)} {newfile}"),
                    input=stdin,
                    text=True,
                )

    def show_diff(self) -> bool:
        newitems = dict(self.newitems)
        olditems = dict(self.olditems)
        if distutils.spawn.find_executable("diff"):
            with tempfile.TemporaryDirectory(prefix="nomad-var-dir_") as tmpd:
                tmpd = Path(tmpd)
                create_tree(tmpd / "nomad", olditems)
                create_tree(tmpd / "local", newitems)
                cmd = "diff --color -ru nomad local"
                log.info(f"+ {cmd}")
                subprocess.run(split(cmd), cwd=tmpd)
        return newitems != olditems


###############################################################################


@click.group(
    help="Given a list of files puts the file content into a nomad variable storage.",
    epilog="Written by Kamil Cukrowski 2023. All right reserved.",
)
@click.option("-n", "--dryrun", is_flag=True)
@click.option("-v", "--verbose", is_flag=True)
@namespace_option()
@click.option("-p", "--path", help="The path of the variable to save")
@click.option(
    "-j",
    "--job",
    help="Equal to --path=nomad/job/<JOB>",
    shell_complete=complete_job,
)
@click.option(
    "-s",
    "--service",
    type=click.Path(exists=True, dir_okay=False),
    help="Get namespace and job name from this nomad service file",
)
@click.option(
    "--disable-size-check",
    is_flag=True,
    help="Disable checking if the file is smaller than 10mb",
)
@click.pass_context
@common_options()
def cli(ctx, **kwargs):
    global args
    args = argparse.Namespace(**ctx.params)


@cli.command(
    "put",
    help="Recursively scan files in given PATHS and upload filenames as key and file content as value to nomad variable store.",
)
@click.option("--force", is_flag=True, help="Like nomad var put -force")
@click.option("--check-index", type=int, help="Like nomad var put -check-index")
@click.option(
    "--relative",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    help="Files have paths relative to this directory instead of current working directory",
)
@click.option(
    "-D",
    "D",
    type=str,
    help="Additional var=value to store in nomad variables",
    multiple=True,
)
@click.option("--clear", is_flag=True, help="Remove keys that are not found in files")
@click.argument(
    "paths",
    type=click.Path(exists=True, path_type=Path),
    nargs=-1,
)
@common_options()
@click.pass_context
def mode_put(ctx, **kvargs):
    nvd = NomadVarDirMain(ctx)
    print(ctx.params)
    if args.force:
        assert args.check_index is None, "either --force or --check-index"
    nvd.gen_new_items()
    if not nvd.show_diff():
        log.info("No difference in var {args.path}@{args.namespace}")
    else:
        keysstr = " ".join(nvd.newitems.keys())
        log.info(
            f"{dryrunstr()}Putting var {args.path}@{args.namespace} with keys: {keysstr}"
        )
        if not args.dryrun:
            try:
                mynomad.variables.create(
                    args.path,
                    nvd.newitems,
                    cas=args.check_index
                    if args.check_index is not None
                    else 0
                    if not args.force
                    else None,
                )
            except VariableConflict as e:
                raise Exception(
                    f"Variable update conflict. Pass --check-index={e.variable.ModifyIndex}"
                ) from e


@cli.command("diff", help="Show only diff")
@click.option(
    "--relative",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    help="Files have paths relative to this directory instead of current working directory",
)
@click.argument(
    "paths",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    nargs=-1,
)
@click.pass_context
@common_options()
def mode_diff(ctx, **kwargs):
    nvd = NomadVarDirMain(ctx)
    nvd.gen_new_items()
    isdiff = nvd.show_diff()
    exit(2 if isdiff else 0)


@cli.command(
    "get",
    help="Get files stored in nomad variables adnd store them in specific directory",
)
@click.argument(
    "dest",
    type=click.Path(file_okay=False, writable=True, path_type=Path),
)
@click.pass_context
@common_options()
def mode_get(ctx, **kwargs):
    nvd = NomadVarDirMain(ctx)
    #
    args.dest.mkdir(exist_ok=True, parents=True)
    for file, content in nvd.olditems.items():
        file = args.dest / file
        if file.exists() and file.is_file() and file.read_text() == content:
            log.info(f"nochange {file}")
        else:
            log.info(f"{dryrunstr()}writting {file}")
            if not args.dryrun:
                file.parent.mkdir(exist_ok=True, parents=True)
                with file.open("w+") as f:
                    f.write(content)


###############################################################################

if __name__ == "__main__":
    cli.main()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-3.0-or-later

import argparse
import logging
from typing import Any, List, Optional, Union

import click.shell_completion

from . import nomadlib
from .common import alias_option, common_options, mynomad, nomad_find_job

log = logging.getLogger(__name__)


def gen_alloc(alloc: nomadlib.Alloc) -> List[str]:
    out = []
    log.debug(f"Parsing {alloc.ID}")
    ports = ((alloc.get("AllocatedResources") or {}).get("Shared") or {}).get("Ports")
    for i in ports or []:
        params = dict(
            **{
                k: v
                for k, v in alloc.items()
                if isinstance(v, str) and isinstance(k, str)
            },
            label=i["Label"],
            host=i["HostIP"],
            port=i["Value"],
        )
        # If label if given, filter.
        if args.label:
            if params["label"] != args.label:
                log.debug(f"Filtered {i}")
                continue
        log.debug(f"Found {i}")
        try:
            txt = args.format.format(**params)
        except KeyError:
            log.exception(f"format={args.format!r} params={params}")
            raise
        out += [txt]
    return out


class IdArgument(click.ParamType):
    name = "id"

    def __is_alloc(self, ctx: Optional[click.Context]):
        return ctx and ctx.params.get("alloc")

    def convert(
        self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Union[nomadlib.Alloc, str]:
        id = value
        if self.__is_alloc(ctx):
            allocs = mynomad.get(f"allocations", params={"prefix": id})
            assert len(allocs) > 0, f"Allocation with id {id} not found"
            assert len(allocs) < 2, f"Multiple allocations found starting with id {id}"
            return nomadlib.Alloc(allocs[0])
        return nomad_find_job(id)

    def shell_complete(
        self, ctx: click.Context, param: click.Parameter, incomplete: str
    ) -> List[click.shell_completion.CompletionItem]:
        arr = (
            mynomad.get("allocations") if self.__is_alloc(ctx) else mynomad.get("jobs")
        )
        ids = [id for id in (x["ID"] for x in arr) if id.startswith(incomplete)]
        return [click.shell_completion.CompletionItem(id) for id in ids]


default_format = "{host}:{port}"
long_format = "{host}:{port} {label} {Name} {ID}"


@click.command(
    help="""
Print dynamic ports allocated by Nomad for a specific job or allocation.
If no ports are found, exit with 2 exit status.
If label argument is given, outputs only redirects that match given label.
"""
)
@click.option(
    "-f",
    "--format",
    default=default_format,
    help=f"The python .format() to print the output with. [default: {default_format!r}]",
)
@alias_option(
    "-l",
    "--long",
    aliased=dict(format=long_format),
)
@click.option("-s", "--separator", default="\n", show_default=True)
@click.option("-v", "--verbose", count=True, help="Be more verbose.")
@click.option("--alloc", is_flag=True, help="The argument is an allocation, not job id")
@click.option(
    "--all",
    help="Show all job allocation ports, not only running or pending allocations.",
    is_flag=True,
)
@common_options()
@click.argument("id", type=IdArgument())
@click.argument("label", required=False)
def cli(id: Union[nomadlib.Alloc, str], **kwargs):
    global args
    args = argparse.Namespace(**kwargs)
    logging.basicConfig(
        level=(
            logging.DEBUG
            if args.verbose > 0
            else logging.INFO
            if args.verbose == 0
            else logging.WARN
            if args.verbose == 1
            else logging.ERROR
        ),
    )
    if args.alloc:
        assert isinstance(id, nomadlib.Alloc)
        alloc = id
        out = gen_alloc(alloc)
    else:
        assert isinstance(id, str)
        jobid = id
        out = []
        for alloc in mynomad.get(f"job/{jobid}/allocations"):
            if args.all or nomadlib.Alloc(alloc).is_pending_or_running():
                # job/*/allocations does not have AllocatedResources information.
                alloc = nomadlib.Alloc(mynomad.get(f"allocation/{alloc['ID']}"))
                out += gen_alloc(alloc)
    if not out:
        exit(2)
    print(*out, sep=args.separator)


if __name__ == "__main__":
    cli()



#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-3.0-or-later

import dataclasses
import json
import logging
import shlex
import shutil
import subprocess
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, TextIO, Tuple, Union

from .common import mynomad

log = logging.getLogger(__name__)

###############################################################################


def nomad_job_run(
    file: str,
    isjson: Optional[bool] = None,
    stdin: Optional[Union[TextIO, int]] = None,
    input: Optional[str] = None,
) -> str:
    """Call nomad job run to start a Nomad job from a file or from stdin or from input"""
    # When file is "-", either stdin or input has to be given.
    # log.debug(f"file={file} isjson={isjson} stdin={stdin} input={input}")
    if file == "-":
        if stdin is None:
            assert input is not None
        else:
            assert input is None
    else:
        assert stdin is None
        assert input is None
    #
    jsonarg = "-json" if isjson else ""
    cmd = shlex.split(f"nomad job run -detach -verbose {jsonarg} {shlex.quote(file)}")
    log.info(f"+ {' '.join(shlex.quote(x) for x in cmd)}")
    try:
        rr = subprocess.run(
            cmd,
            stdin=stdin
            if stdin is not None
            else None
            if input is not None
            else subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            input=input,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        # nomad will print its error, we can just exit
        exit(e.returncode)
    # Extract evaluation id rfom Nomad output.
    data = rr.stdout.strip()
    for line in data.splitlines():
        log.info(line)
    evalid = next(
        x.split(" ", 3)[-1] for x in data.splitlines() if "eval" in x.lower()
    ).strip()
    return evalid


###############################################################################


@dataclasses.dataclass
class Detail(ABC):
    # The actual inputed string from the user.
    input: str
    # This is true, we know that content is a JSON.
    forcejson: bool

    @abstractmethod
    def seekable_read(self) -> Optional[str]:
        """Read the content of the file only if it is seekable."""
        raise NotImplementedError()

    def read(self) -> str:
        """Return the content of the file"""
        content = self.seekable_read()
        assert content is not None
        return content

    @abstractmethod
    def nomad_start(self) -> str:
        raise NotImplementedError()

    def close(self):
        """Call on finally to close potentially opened file"""
        pass

    def seekable_isjson(self):
        """If the underlying file is seekable, check if it is a json"""
        if self.forcejson:
            return True
        content = self.seekable_read()
        if content is not None:
            # Detect if its a json if we can.
            try:
                json.loads(content)
                return True
            except json.JSONDecodeError:
                return False
        return None

    def nomad_start_in(
        self,
        file: str,
        stdin: Optional[Union[TextIO, int]] = None,
        input: Optional[str] = None,
    ):
        return nomad_job_run(
            file=file,
            isjson=self.seekable_isjson(),
            stdin=stdin,
            input=input,
        )

    def read_dict(self) -> Tuple[str, dict]:
        content = self.read()
        # Assume input is a string with valid job specification.
        try:
            # try json
            data = json.loads(content)
        except json.JSONDecodeError:
            if self.forcejson:
                raise
            data = json.loads(mynomad.jobhcl2json(content))
        return content, data

    def api_start(self):
        content, data = self.read_dict()
        evalid = mynomad.start_job(data, content)["EvalID"]
        return evalid

    def run(self) -> str:
        return self.nomad_start() if shutil.which("nomad") else self.api_start()


@dataclasses.dataclass
class ModeString(Detail):
    def seekable_read(self):
        return self.input

    def nomad_start(self):
        return self.nomad_start_in(file="-", input=self.input)


class ModeFilehandle(Detail):
    @abstractmethod
    def filehandle(self):
        raise NotImplementedError()

    def seekable_read(self):
        filehandle = self.filehandle()
        if filehandle.seekable():
            content = filehandle.read()
            filehandle.seek(0)
            return content
        return None

    def read(self):
        return self.filehandle().read()


class ModeStdin(ModeFilehandle):
    def filehandle(self):
        return sys.stdin

    def nomad_start(self):
        return self.nomad_start_in(file="-", stdin=sys.stdin)


@dataclasses.dataclass
class ModeFile(ModeFilehandle):
    file: Path
    _filehandle: Optional[TextIO] = None

    def filehandle(self):
        if self._filehandle is None:
            self._filehandle = self.file.open()
        return self._filehandle

    def close(self):
        if self._filehandle is not None:
            self._filehandle.close()

    def nomad_start(self):
        return self.nomad_start_in(file=str(self.file))


def nomad_smart_start_job(input: str, forcejson: bool = False) -> str:
    """
    Be "smartly" by detecting if input is a file or not and run it depending on that.
    If input contains a newline, it means it is a JSON or HCL job specification.
    If input is the string '-', it means we run from standard input.
    If input does not have a newline nor is '-', the input is a filename.
    If input is standard input or a filename, if it is seekable and forcejson is false,
    then read the input and if it is a valid json, parse it as a JSON.
    Otherwise parse it as a HCL file.
    If Nomad executable is available, run the job with Nomad executable.
    Otherwise use the API.
    """
    mode = (
        ModeStdin(input, forcejson)
        if input == "-"
        else ModeString(input, forcejson)
        if "\n" in input or (input.count("{") > 5 and input.count("}") > 5)
        else ModeFile(input, forcejson, Path(input))
    )
    try:
        return mode.run()
    finally:
        mode.close()



import datetime

from .connection import *
from .types import *


def escape(txt: str) -> str:
    # Protect against "interpolation expression" by nomad.
    # See https://www.nomadproject.io/docs/job-specification/hcl2/expressions#string-templates
    return txt.replace("%{", "%%{").replace("${", "$${")


def unescape(txt: str) -> str:
    # See nomad_escape
    return txt.replace("%%{", "%{").replace("$${", "${")


def ns2s(ns: int) -> float:
    """Convert nanoseconds to float seconds"""
    return ns / 1000000000


def ns2dt(ns: int) -> datetime.datetime:
    """Convert nanoseconds to datetime"""
    return datetime.datetime.fromtimestamp(ns2s(ns)).astimezone()



from __future__ import annotations

import copy
import enum
from typing import Any, ChainMap, Type, Union, get_type_hints


def all_annotations(cls) -> ChainMap[str, Type]:
    """
    Returns a dictionary-like ChainMap that includes annotations for all
    attributes defined in cls or inherited from superclasses.
    Also resolve runtime type hints - https://peps.python.org/pep-0563/
    """
    return ChainMap(*(get_type_hints(c) for c in cls.__mro__))


def _init_value(classname: str, dstname: str, dsttype: Any, srcval: Any):
    # print(f"Constructing {aname} with type {atype} from {val}")
    dstorigin = getattr(dsttype, "__origin__", None)

    def msg() -> str:
        return (
            f"Error when constructing class {classname!r} expected type {dsttype!r} with origin {dstorigin!r}"
            f" for field {dstname!r}, but received {type(srcval)} with value: {srcval!r}"
        )

    try:
        if dstorigin == list:
            assert type(srcval) == dstorigin, msg()
            return [
                _init_value(classname, dstname, dsttype.__args__[0], x) for x in srcval
            ]
        elif dstorigin == dict:
            assert type(srcval) == dstorigin, msg()
            return {
                _init_value(classname, dstname, dsttype.__args__[0], k): _init_value(
                    classname, dstname, dsttype.__args__[1], v
                )
                for k, v in srcval.items()
            }
        elif dstorigin == Union:
            if type(srcval) in dsttype.__args__:
                return srcval
            return _init_value(classname, dstname, dsttype.__args__[0], srcval)
        elif dsttype == Any:
            return srcval
        elif issubclass(dsttype, DataDict):
            return dsttype(srcval)
        elif issubclass(dsttype, enum.Enum):
            return dsttype(srcval)
        return srcval
    except AssertionError:
        raise
    except Exception:
        raise Exception(msg())


def _asdict_value(fname: str, val: Any):
    if isinstance(val, list):
        return [_asdict_value(fname, x) for x in val]
    elif isinstance(val, dict):
        return {
            _asdict_value(fname, k): _asdict_value(fname, v) for k, v in val.items()
        }
    elif isinstance(val, enum.Enum):
        return val.value
    elif hasattr(val, fname):
        return getattr(val, fname)()
    return val


class DataDict:
    """
    Data dictionary - a mix between dataclass and AttrDict from stackoverflow.
    Basically a dictionary with standard dictionary accesseses proxied to self.__dict__.
    Then additional __init__ function that allows constructing from a dictionary and will
    also construct any nested AttrDict objects from type hints.
    """

    def __init__(self, *args, **kwargs):
        data = dict(*args, **kwargs)
        annotations = all_annotations(self.__class__)
        # Copy default values from class.
        for akey, atype in annotations.items():
            if akey in self.__class__.__dict__:
                self.__dict__[akey] = copy.deepcopy(self.__class__.__dict__[akey])
        # Intialize values from dictionary.
        for key, val in data.items():
            if key in annotations:
                self.__dict__[key] = _init_value(
                    self.__class__.__name__, key, annotations[key], val
                )
            else:
                self.__dict__[key] = val
        self.__post_init__()

    def __post_init__(self):
        pass

    def __setitem__(self, k, v):
        self.__dict__[k] = v

    def __getitem__(self, k):
        return self.__dict__[k]

    def get(self, k, v=None):
        return self.__dict__.get(k, v)

    def setdefault(self, k, v):
        return self.__dict__.setdefault(k, v)

    def __contains__(self, k):
        return k in self.__dict__

    def __delitem__(self, k):
        del self.__dict__[k]

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def asdict(self):
        """Convert datadict to a dictionary"""
        fname = self.asdict.__name__
        return {k: _asdict_value(fname, v) for k, v in self.__dict__.items()}

    def update(self, o: Union[DataDict, dict]):
        """Update values with other values"""
        if isinstance(o, DataDict):
            self.__dict__.update(o)
        else:
            self.__dict__.update(self.__class__(o))
        return self

    def remove_none(self, recursive=True):
        """Remove all elements that are set to None"""
        self.__dict__ = {
            key: val.remove_none() if recursive and isinstance(val, DataDict) else val
            for key, val in self.__dict__.items()
            if val is not None
        }
        return self

    def __repr__(self):
        data = " ".join(f"{k}={self.__dict__[k]!r}" for k in sorted(self.__dict__))
        return f"{self.__class__.__name__}({data})"

    def __eq__(self, o):
        if self.__class__ == o.__class__:
            return self.__dict__ == o.__dict__
        else:
            return False



import dataclasses
import subprocess
import sys


@dataclasses.dataclass
class Colors:
    """ASCII TERM colors"""

    bold: str = "bold"
    reset: str = "sgr0"
    black: str = "setaf 0"
    red: str = "setaf 1"
    green: str = "setaf 2"
    orange: str = "setaf 3"
    blue: str = "setaf 4"
    magenta: str = "setaf 5"
    cyan: str = "setaf 6"
    white: str = "setaf 7"
    brightblack: str = "setaf 8"
    brightred: str = "setaf 9"
    brightgreen: str = "setaf 10"
    brightorange: str = "setaf 11"
    brightblue: str = "setaf 12"
    brightmagenta: str = "setaf 13"
    brightcyan: str = "setaf 14"
    brightwhite: str = "setaf 15"


empty = Colors(**{f.name: "" for f in dataclasses.fields(Colors)})
"""Empty colors"""


def init_ex() -> Colors:
    """Return Colors with ANSI escape sequences extracted from tput"""
    if not sys.stdout.isatty() or not sys.stderr.isatty():
        return empty
    tputdict = dataclasses.asdict(Colors())
    tputscript = "\n".join(tputdict.values()).replace("\n", "\nlongname\nlongname\n")
    longname = subprocess.run(
        f"tput longname".split(),
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    ).stdout
    # Protect against empty longname.
    if not longname:
        return empty
    ret = subprocess.run(
        "tput -S".split(),
        input=tputscript,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
    ).stdout
    retarr = ret.split(f"{longname}{longname}")
    if len(tputdict.keys()) != len(retarr):
        return empty
    return Colors(**{k: v for k, v in zip(tputdict.keys(), retarr)})


def init() -> Colors:
    try:
        return init_ex()
    except Exception:
        return empty



import dataclasses
import functools
import inspect
import json
import os
import re
import shlex
import subprocess
import sys
import tempfile
import time
from pathlib import Path
from typing import IO, Dict, List, Optional, Union

from nomad_tools import nomadlib

os.environ.setdefault("NOMAD_NAMESPACE", "default")


def caller(up=0):
    return inspect.stack()[1 + up][3]


def job_exists(jobname):
    try:
        nomadlib.NomadConn().get(f"job/{jobname}")
        return True
    except nomadlib.JobNotFound:
        return False


@functools.lru_cache(maxsize=0)
def nomad_has_docker():
    nodes = json.loads(subprocess.check_output("nomad node status -json".split()))
    for n in nodes:
        if n["Drivers"].get("docker", {}).get("Healthy"):
            return True
    return False


@dataclasses.dataclass
class NamedTemporaryFileContent:
    content: str
    suffix: Optional[str] = None
    file: Optional[IO] = None

    def __enter__(self):
        self.file = tempfile.NamedTemporaryFile("w", suffix=self.suffix)
        self.file.write(self.content)
        self.file.flush()
        return self.file.name

    def __exit__(self, *args):
        if self.file:
            self.file.close()


def gen_job(
    script=""" echo hello world """, name: Optional[str] = None, mode: str = "raw_exec"
):
    assert name is None or " " not in name
    name = (
        name
        or os.environ.get("PYTEST_CURRENT_TEST", caller(1)).split(":")[-1].split(" ")[0]
    )
    jobname = f"test-nomad-utils-{name}"
    docker_task = {
        "Driver": "docker",
        "Config": {
            "image": "busybox:stable",
            "command": "sh",
            "args": ["-xc", script],
            "init": True,
        },
    }
    raw_exec_task = {
        "Driver": "raw_exec",
        "Config": {
            "command": "sh",
            "args": ["-xc", script],
        },
    }
    task = raw_exec_task if mode == "raw_exec" else docker_task
    job = {
        "ID": jobname,
        "Type": "batch",
        "Meta": {
            "TIME": f"{time.time_ns()}",
        },
        "TaskGroups": [
            {
                "Name": jobname,
                "ReschedulePolicy": {"Attempts": 0},
                "RestartPolicy": {"Attempts": 0},
                "Tasks": [
                    {
                        **task,
                        "Name": jobname,
                    }
                ],
            }
        ],
    }
    return {"Job": job}


def get_testjobs() -> Dict[str, Path]:
    files = list(Path(f"./jobs/").glob("*.nomad.hcl"))
    assert len(files) > 3
    ret = {f.name[: -len(".nomad.hcl")]: f for f in files}
    print(ret)
    return ret


def quotearr(cmd: List[str]):
    return " ".join(shlex.quote(x) for x in cmd)


def run(
    cmd: str,
    check: Optional[Union[bool, int, List[int]]] = True,
    text=True,
    stdout: Union[bool, int] = False,
    output: Union[List[str], List[Union[str, re.Pattern]], List[re.Pattern]] = [],
    input: Optional[str] = None,
    **kwargs,
) -> subprocess.CompletedProcess:
    cmda: List[str] = []
    for i in shlex.split(cmd):
        if i in "nomad-cp nomad-watch nomad-port".split():
            cmda.extend(f"python3 -m nomad_tools.{i.replace('-', '_')} -v".split())
        else:
            cmda.append(i)
    print(" ", file=sys.stderr, flush=True)
    print(f"+ {quotearr(cmda)}", file=sys.stderr, flush=True)
    stdout = stdout or bool(output)
    # Run subprocess.Popen, input stdin and output stdout.
    with subprocess.Popen(
        cmda,
        text=text,
        stdout=subprocess.PIPE if stdout else sys.stderr,
        stdin=subprocess.PIPE if input else subprocess.DEVNULL,
        **kwargs,
    ) as pp:
        if input:
            assert pp.stdin
            pp.stdin.write(input)
            pp.stdin.close()
        captured_stdout = None
        if stdout:
            captured_stdout = ""
            assert pp.stdout
            for line in pp.stdout:
                line = line.rstrip()
                captured_stdout += line + "\n"
                print(line)
    rr = subprocess.CompletedProcess(cmda, pp.returncode, captured_stdout, None)
    #
    if check is True:
        check = [0]
    if check is None:
        check = None
    elif check is False:
        assert (
            rr.returncode != 0
        ), f"Command {rr} died with zero {rr.returncode}, which is wrong"
        check = None
    if isinstance(check, int):
        check = [check]
    if isinstance(check, list):
        assert (
            rr.returncode in check
        ), f"Command {rr} died with {rr.returncode} not in {check}"
    #
    if output:
        assert rr.stdout is not None
        for pat in output:
            if isinstance(pat, str):
                assert pat in rr.stdout, f"String {pat} not found"
            elif isinstance(pat, re.Pattern):
                assert pat.findall(rr.stdout.strip()), f"Pattern {pat} not found"
            else:
                assert 0
    #
    return rr


def run_nomad_cp(cmd: str, **kwargs):
    return run(f"python3 -m nomad_tools.nomad_cp -v {cmd}", **kwargs)


def run_nomad_watch(
    cmd: str,
    pre: str = "",
    check: Optional[Union[bool, int, List[int]]] = True,
    text=True,
    stdout: Union[bool, int] = False,
    output: Union[List[str], List[Union[str, re.Pattern]], List[re.Pattern]] = [],
    input: Optional[str] = None,
    **kwargs,
):
    return run(
        f"{pre} python3 -m nomad_tools.nomad_watch -v {cmd}",
        check=check,
        text=text,
        stdout=stdout,
        output=output,
        input=input,
        **kwargs,
    )


def run_nomad_vardir(cmd: str, **kwargs):
    return run(f"python3 -m nomad_tools.nomad_vardir -v {cmd}", **kwargs)
